"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.iam.StatementResources import StatementResources, _Arn


@pytest.fixture(scope="module")
def rule():
    rule = StatementResources()
    yield rule


@pytest.fixture
def template():
    return {
        "Parameters": {
            "MyParameter": {"Type": "String"},
        },
        "Resources": {
            "LogGroup": {
                "Type": "AWS::Logs::LogGroup",
            }
        },
    }


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            [],  # wrong type
            [],
        ),
        (
            {
                "Action": "cloudformation:CreateStackSet",
                "Resource": "*",
            },
            [],
        ),
        (
            {
                "Action": ["cloudformation:CreateStackSet"],
                "Resource": ["*"],
            },
            [],
        ),
        (
            {
                "Action": ["cloudformation:CreateStack"],
                "Resource": "*",
            },
            [],
        ),
        (
            # action with no colon
            {
                "Action": ["cloudformationCreateStackSet"],
                "Resource": ["arn:aws:cloudformation:us-east-1:123456789012:*"],
            },
            [],
        ),
        (
            # service not in map
            {
                "Action": ["foo:CreateStackSet"],
                "Resource": ["arn:aws:cloudformation:us-east-1:123456789012:*"],
            },
            [],
        ),
        (
            # action not in map
            {
                "Action": ["cloudformation:foo"],
                "Resource": ["arn:aws:cloudformation:us-east-1:123456789012:*"],
            },
            [],
        ),
        (
            # skip validation on asterisk actions
            {
                "Action": ["foo:Create*"],
                "Resource": ["arn:aws:cloudformation:us-east-1:123456789012:*"],
            },
            [],
        ),
        (
            {
                "Action": [{"Ref": "MyParameter"}],
                "Resource": ["arn:aws:cloudformation:us-east-1:123456789012:*"],
            },
            [],
        ),
        (
            {
                "Action": {"Ref": "MyParameter"},
                "Resource": ["arn:aws:cloudformation:us-east-1:123456789012:*"],
            },
            [],
        ),
        (
            {
                "Action": "cloudformation:CreateStackSet",
                "Resource": [{"Ref": "MyParameter"}],
            },
            [],
        ),
        (
            {
                "Action": "ec2:CreateTags",
                "Resource": ["arn:aws:ec2:*::snapshot/*"],
            },
            [],
        ),
        (
            {
                "Action": "cloudformation:CreateStackSet",
                "Resource": [{"Ref": "LogGroup"}],
            },
            [
                # ValidationError(
                #    "action 'cloudformation:CreateStackSet' requires a resource of '*'",  # noqa: E501
                #    rule=StatementResources(),
                #    path=deque(["Resource"]),
                # ),
            ],
        ),
        (
            {
                "Action": "cloudformation:CreateStackSet",
                "Resource": [{"Foo": "Bar"}],
            },
            [
                # ValidationError(
                #    "action 'cloudformation:CreateStackSet' requires a resource of '*'",  # noqa: E501
                #    rule=StatementResources(),
                #    path=deque(["Resource"]),
                # ),
            ],
        ),
        (
            {
                "Action": "cloudformation:CreateStackSet",
                "Resource": ["arn:aws:cloudformation:us-east-1:123456789012:*"],
            },
            [
                # ValidationError(
                #    "action 'cloudformation:CreateStackSet' requires a resource of '*'",  # noqa: E501
                #    rule=StatementResources(),
                #    path=deque(["Resource"]),
                # ),
            ],
        ),
        (
            {
                "Action": [
                    "cloudformation:CreateStackSet",
                    {"Ref": "MyParameter"},
                ],
                "Resource": ["arn:aws:cloudformation:us-east-1:123456789012:*"],
            },
            [
                # ValidationError(
                #    "action 'cloudformation:CreateStackSet' requires a resource of '*'",  # noqa: E501
                #    rule=StatementResources(),
                #    path=deque(["Resource"]),
                # ),
            ],
        ),
        (
            {
                "Action": ["cloudformation:CreateStackSet"],
                "Resource": ["arn:aws:cloudformation:us-east-1:123456789012:*"],
            },
            [
                # ValidationError(
                #    "action 'cloudformation:CreateStackSet' requires a resource of '*'",  # noqa: E501
                #    rule=StatementResources(),
                #    path=deque(["Resource"]),
                # ),
            ],
        ),
        (
            {
                "Action": ["cloudformation:CreateStackSet"],
                "Resource": ["arn:aws:cloudformation:us-east-1:123456789012:*", "*"],
            },
            [],
        ),
        (
            {
                "Action": [
                    "cloudformation:CreateStackSet",
                    "cloudformation:CreateStack",
                ],
                "Resource": ["arn:aws:cloudformation:*:*:stack/*"],
            },
            [
                # ValidationError(
                #    "action 'cloudformation:CreateStackSet' requires a resource of '*'",  # noqa: E501
                #    rule=StatementResources(),
                #    path=deque(["Resource"]),
                # ),
            ],
        ),
        (
            {
                "Action": ["cloudformation:CreateStack"],
                "Resource": ["arn:aws:cloudformation:us-east-1:123456789012:stack/*"],
            },
            [],
        ),
        (
            {
                "Action": ["cloudformation:CreateStack"],
                "Resource": ["arn:aws:cloudformation:us-east-1:123456789012:*"],
            },
            [],
        ),
        (
            {
                "Action": ["cloudformation:CreateStack"],
                "Resource": [
                    "arn:aws:cloudformation:us-east-1:123456789012:stack/dne/*"
                ],
            },
            [],
        ),
        (
            {
                "Action": ["cloudformation:CreateStack"],
                "Resource": ["arn:aws:cloudformation:us-east-1:123456789012:*dne*"],
            },
            [],
        ),
        (
            {
                "Action": ["cloudformation:CreateStack"],
                "Resource": ["arn:aws:logs:us-east-1:123456789012:stack/dne/*"],
            },
            [
                ValidationError(
                    (
                        "action 'cloudformation:CreateStack' "
                        "requires a resource of "
                        "['arn:${Partition}:cloudformation:${Region}:${Account}:stack/.*']"
                    ),
                    rule=StatementResources(),
                    path=deque(["Resource"]),
                ),
            ],
        ),
        (
            {
                "Action": ["cloudformation:Create?tack"],
                "Resource": ["arn:aws:logs:us-east-1:123456789012:stack/dne/*"],
            },
            [],
        ),
        (
            {
                "Action": ["cloudformation:CreateStack"],
                "Resource": ["arn:aws:logs:us-east-1:123456789012:stack/dne/*"],
            },
            [
                ValidationError(
                    (
                        "action 'cloudformation:CreateStack' "
                        "requires a resource of "
                        "['arn:${Partition}:cloudformation:${Region}:${Account}:stack/.*']"
                    ),
                    rule=StatementResources(),
                    path=deque(["Resource"]),
                ),
            ],
        ),
        (
            {
                "Action": "logs:CreateLogGroup",
                "Resource": "arn:aws:logs:us-east-1:123456789012:log-group:dne",
            },
            [],
        ),
        (
            {
                "Action": "logs:CreateLogGroup",
                "Resource": "arn:aws:logs:us-east-1:123456789012:log-group*",
            },
            [],
        ),
        (
            {
                "Action": [
                    "cloudformation:CreateStack",
                    "logs:CreateLogGroup",
                ],
                "Resource": [
                    "arn:aws:cloudformation:us-east-1:123456789012:stack/dne/*",
                    "arn:aws:logs:us-east-1:123456789012:log-group:dne",
                ],
            },
            [],
        ),
        (
            # no delimiter in resource
            {
                "Action": [
                    "codepipeline:putapprovalresult",
                ],
                "Resource": [
                    "arn:aws:codepipeline:us-east-1:123456789012:resource-name",
                ],
            },
            [],
        ),
        (
            {
                "Action": [
                    "cloudformation:CreateStack",
                    "logs:CreateLogGroup",
                ],
                "Resource": [
                    "arn:aws:cloudformation:us-east-1:123456789012:changeSet/dne/*",
                    "arn:aws:logs:us-east-1:123456789012:log-group:dne",
                ],
            },
            [
                ValidationError(
                    (
                        "action 'cloudformation:CreateStack' requires "
                        "a resource of "
                        "['arn:${Partition}:cloudformation:${Region}:${Account}:stack/.*']"
                    ),
                    rule=StatementResources(),
                    path=deque(["Resource"]),
                ),
            ],
        ),
        (
            # should not fail on the actions that require a resource
            # but still fails because there is no asterisk
            {
                "Action": [
                    "logs:CreateLogStream",
                    "logs:PutLogEvents",
                    "logs:DescribeLogGroups",
                    "logs:DescribeLogStreams",
                ],
                "Resource": {"Fn::GetAtt": "LogGroup.Arn"},
            },
            [
                # ValidationError(
                #    "action 'logs:DescribeLogGroups' requires a resource of '*'",  # noqa: E501
                #    rule=StatementResources(),
                #    path=deque(["Resource"]),
                # ),
            ],
        ),
    ],
)
def test_rule(instance, expected, rule, validator):
    errors = list(rule.validate(validator, {}, instance, {}))

    assert errors == expected, f"Got {errors!r}"


def test_added():
    arn = _Arn("arn:aws:cloudformation:us-east-1:123456789012:stack/*")

    assert arn.__repr__() == "arn:aws:cloudformation:us-east-1:123456789012:stack/*"
    assert arn != 1

    greengress_1 = _Arn(
        "arn:aws:cloudformation:us-east-1:123456789012:/greengrass/definition/connectors/.*"
    )
    greengress_2 = _Arn("arn:aws:cloudformation:us-east-1:123456789012:/greengrass")
    greengrass_3 = _Arn(
        "arn:aws:cloudformation:us-east-1:123456789012:/greengrass/things/"
    )

    assert greengress_1 == greengress_2
    assert greengress_1 != greengrass_3

    assert _Arn("arn:aws:cloudformation:us-east-1:123456789012:") == _Arn(
        "arn:aws:cloudformation:us-east-1:123456789012:"
    )
    assert _Arn("arn:aws:cloudformation:us-east-1:123456789012:test/") == _Arn(
        "arn:aws:cloudformation:us-east-1:123456789012:test/"
    )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.data.schemas.extensions.aws_route53_recordset
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class RecordSet(CfnLintJsonSchema):
    """Check Route53 Recordset Configuration"""

    id = "E3023"
    shortdesc = "Validate Route53 RecordSets"
    description = "Check if all RecordSets are correctly configured"
    source_url = "https://docs.aws.amazon.com/Route53/latest/DeveloperGuide/ResourceRecordTypes.html"
    tags = ["resources", "route53", "record_set"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::Route53::RecordSet/Properties",
                "Resources/AWS::Route53::RecordSetGroup/Properties/RecordSets/*",
            ],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_route53_recordset,
                filename="recordset_pattern.json",
            ),
            all_matches=True,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from test.unit.rules import BaseRuleTestCase

from cfnlint.context import Context, Path
from cfnlint.jsonschema import CfnTemplateValidator
from cfnlint.rules.parameters.ValuePattern import ValuePattern
from cfnlint.template.template import Template


class TestValuePattern(BaseRuleTestCase):
    """Test Allowed Value Parameter Configuration"""

    def setUp(self):
        """Setup"""
        self.rule = ValuePattern()
        cfn = Template(
            "test.yaml",
            {
                "Parameters": {
                    "1": {"Type": "String", "Default": "A"},
                    "2": {"Type": "String", "Default": 1},
                    "3": {
                        "Type": "String",
                        "AllowedValues": [
                            "A",
                            "B",
                        ],
                    },
                    "4": {
                        "Type": "String",
                        "AllowedValues": [
                            1,
                            2,
                        ],
                    },
                    "5": [],
                    "6": {"Type": "String", "AllowedValues": {"foo": "bar"}},
                    "7": {"Type": "String", "AllowedValues": [{"foo": "bar"}]},
                }
            },
            regions=["us-east-1"],
        )
        self.rule.initialize(cfn)

    def test_validate(self):
        validator = CfnTemplateValidator(
            schema={"type": "string"},
            context=Context(
                "us-east-1",
                {},
                path=Path(
                    path=deque(["Ref", "MyParameter"]),
                    value_path=deque(["Parameters", "MyParameter", "Default"]),
                ),
            ),
        )

        errs = list(self.rule.pattern(validator, "^[A-Z]$", "1", {}))
        self.assertEqual(len(errs), 1)
        for err in errs:
            self.assertEqual(err.rule.id, "W2031")
            self.assertEqual(err.message, "'1' does not match '^[A-Z]$'")
            self.assertEqual(err.rule, self.rule)
            self.assertEqual(
                err.path_override, deque(["Parameters", "MyParameter", "Default"])
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class DomainValidationOptions(CfnLintKeyword):
    """Check if a certificate's domain validation options are set up correctly"""

    id = "E3503"
    shortdesc = "ValidationDomain is superdomain of DomainName"
    description = (
        "In ValidationDomainOptions, the ValidationDomain must be a superdomain of the"
        " DomainName being validated"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-certificatemanager-certificate-domainvalidationoption.html#cfn-certificatemanager-certificate-domainvalidationoption-validationdomain"
    tags = [
        "certificate",
        "certificatemanager",
        "domainvalidationoptions",
        "validationdomain",
    ]

    def __init__(self):
        """Init"""
        super().__init__(
            [
                "Resources/AWS::CertificateManager::Certificate/Properties/DomainValidationOptions/*"
            ]
        )

    def validate(
        self, validator: Validator, _, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        if not isinstance(instance, dict):
            return

        property_sets = validator.cfn.get_object_without_conditions(instance)
        for property_set in property_sets:
            properties = property_set.get("Object")
            domain_name = properties.get("DomainName", None)
            validation_domain = properties.get("ValidationDomain", None)
            if isinstance(domain_name, str) and isinstance(validation_domain, str):
                if domain_name == validation_domain:
                    continue

                if not domain_name.endswith("." + validation_domain):
                    yield ValidationError(
                        (
                            f"{validation_domain!r} must be a "
                            f"superdomain of {domain_name!r}"
                        ),
                        path=deque(["DomainName"]),
                        instance=domain_name,
                    )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import regex as re

from cfnlint.jsonschema import Validator
from cfnlint.rules import CloudFormationLintRule


class FormatKeyword(CloudFormationLintRule):

    def __init__(self, format: str | None = None, pattern: str | None = None) -> None:
        super().__init__()
        self.format_keyword = format
        self.parent_rules = ["E1103"]
        self.pattern = pattern or r""

    def format(self, validator: Validator, instance: Any) -> bool:
        if not isinstance(instance, str):
            return True

        if re.match(self.pattern, instance):
            return True

        return False



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.unit.rules import BaseRuleTestCase

from cfnlint.rules.conditions.Used import Used  # pylint: disable=E0401


class TestUsedConditions(BaseRuleTestCase):
    """Test template mapping configurations"""

    def setUp(self):
        """Setup"""
        super(TestUsedConditions, self).setUp()
        self.collection.register(Used())

    success_templates = [
        "test/fixtures/templates/good/generic.yaml",
    ]

    def test_file_positive(self):
        """Test Positive"""
        self.helper_file_positive()

    def test_file_negative(self):
        """Test failure"""
        self.helper_file_negative("test/fixtures/templates/bad/conditions.yaml", 5)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.data.schemas.other.metadata
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class Configuration(CfnLintJsonSchema):
    """Check if Metadata configuration is properly configured"""

    id = "E4002"
    shortdesc = "Validate the configuration of the Metadata section"
    description = "Validates that Metadata section is an object and has no null values"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/metadata-section-structure.html"
    tags = ["metadata"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Metadata"],
            schema_details=SchemaDetails(
                cfnlint.data.schemas.other.metadata, "configuration.json"
            ),
            all_matches=True,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.ecs.FargateDeploymentSchedulingStrategy import (
    FargateDeploymentSchedulingStrategy,
)


@pytest.fixture(scope="module")
def rule():
    rule = FargateDeploymentSchedulingStrategy()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "LaunchType": "FARGATE",
            },
            [],
        ),
        (
            {
                "LaunchType": "FARGATE",
                "SchedulingStrategy": "REPLICA",
            },
            [],
        ),
        (
            [],  # wrong type
            [],
        ),
        (
            {
                "LaunchType": "FARGATE",
                "SchedulingStrategy": {"Ref": "Replica"},
            },
            [],
        ),
        (
            {
                "LaunchType": "EXTERNAL",
                "SchedulingStrategy": "DAEMON",
            },
            [],
        ),
        (
            {
                "LaunchType": "EXTERNAL",
                "SchedulingStrategy": "DAEMON",
                "DeploymentController": {"Type": "ECS"},
            },
            [],
        ),
        (
            {
                "LaunchType": "EXTERNAL",
                "SchedulingStrategy": "DAEMON",
                "DeploymentController": {"Type": "CODE_DEPLOY"},
            },
            [
                ValidationError(
                    "'REPLICA' was expected",
                    rule=FargateDeploymentSchedulingStrategy(),
                    path=deque(["SchedulingStrategy"]),
                    validator="const",
                    schema_path=deque(
                        ["then", "properties", "SchedulingStrategy", "const"]
                    ),
                )
            ],
        ),
        (
            {
                "LaunchType": "FARGATE",
                "SchedulingStrategy": "DAEMON",
            },
            [
                ValidationError(
                    "'REPLICA' was expected",
                    rule=FargateDeploymentSchedulingStrategy(),
                    path=deque(["SchedulingStrategy"]),
                    validator="const",
                    schema_path=deque(
                        ["then", "properties", "SchedulingStrategy", "const"]
                    ),
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

import regex as re

from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.rules import CloudFormationLintRule


class DynamicReferenceSecretsManagerArn(CloudFormationLintRule):
    id = "W1051"
    shortdesc = (
        "Validate dynamic references to secrets manager "
        "are not used when a secrets manager ARN was expected"
    )
    description = (
        "Certain properties expect a secret manager ARN. This rule "
        "validates if you may be accidently using a secret in place "
        "of the ARN"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/dynamic-references.html#dynamic-references-secretsmanager"
    tags = ["functions", "dynamic reference"]

    def validate(self, validator: Validator, s: Any, instance: Any, schema: Any):

        if "Fn::Sub" == validator.context.path.path[-1]:
            if not re.match(r"^(\\\")?{{resolve:secretsmanager:.*}}(\\\")?$", instance):
                return

        if len(validator.context.path.path) < 3:
            return

        if (
            validator.context.path.path[0] != "Resources"
            or validator.context.path.path[2] != "Properties"
        ):
            return

        fields = [
            "SecretArn",
            "SecretARN",
            "SecretsManagerSecretId",
            "SecretsManagerOracleAsmSecretId",
            "SecretsManagerSecurityDbEncryptionSecretId",
            "SecretsManagerConfiguration",
        ]

        for field in fields:
            if any(field == p for p in validator.context.path.path):
                yield ValidationError(
                    (
                        f"Dynamic reference {instance!r} to secrets manager when "
                        f"the field {field!r} expects the ARN to the secret and "
                        "not the secret"
                    ),
                    rule=self,
                )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from cfnlint.schema import Schema


class Ref:

    def __init__(self, schema: "Schema") -> None:
        primary_ids = schema.schema.get("primaryIdentifier", [])
        if len(primary_ids) > 1:
            self._ref = {"type": "string"}
        else:
            for primary_id in primary_ids:
                self._ref = schema.resolver.resolve_cfn_pointer(primary_id)

    @property
    def ref(self) -> dict[str, Any]:
        return self._ref



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class PipelineFirstStageHasSource(CfnLintKeyword):
    id = "E3700"
    shortdesc = "Validate CodePipeline Source actions are only in the first stage"
    description = (
        "When using AWS::CodePipeline::Pipeline this rule will validate "
        "that Source actions are only used in the first stage"
    )
    tags = ["resources", "codepipeline"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::CodePipeline::Pipeline/Properties/Stages/*/Actions/*/ActionTypeId/Category"
            ],
        )
        self._valid_categories = [
            # removed because Source is only in the first stage
            # "Source",
            "Build",
            "Approval",
            "Deploy",
            "Test",
            "Invoke",
            "Compute",
        ]

    def _is_first_stage(self, full_path: deque) -> bool | None:
        path = list(full_path)[3:]
        i = path.index("Stages")
        while i < len(path):
            if path[i] == "Fn::If":
                i += 2
                continue
            if path[i] == 0:
                return True
            elif isinstance(path[i], int):
                return False
            i += 1

        return None

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        if not validator.is_type(instance, "string"):
            return

        is_first_stage = self._is_first_stage(validator.context.path.path)
        if is_first_stage is None:
            return
        if is_first_stage:
            if instance != "Source":
                yield ValidationError(
                    f"{instance!r} is not one of {['Source']!r}",
                    rule=self,
                    validator="enum",
                )
        else:
            if instance not in self._valid_categories:
                yield ValidationError(
                    f"{instance!r} is not one of {self._valid_categories!r}",
                    rule=self,
                    validator="enum",
                )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.functions._BaseFn import BaseFn


class Length(BaseFn):
    """Check if Length values are correct"""

    id = "E1030"
    shortdesc = "Length validation of parameters"
    description = "Making sure Fn::Length is configured correctly"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-length.html"
    tags = ["functions", "length"]

    def __init__(self) -> None:
        super().__init__(
            "Fn::Length",
            ("integer",),
        )

    def fn_length(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> ValidationResult:
        if not validator.context.transforms.has_language_extensions_transform():
            yield ValidationError(
                (
                    f"{self.fn.name} is not supported without "
                    "'AWS::LanguageExtensions' transform"
                ),
                validator=self.fn.py,
                rule=self,
            )
            return

        for err in super().validate(validator, s, instance, schema):
            err.rule = self
            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from cfnlint.rules.functions._BaseFn import BaseFn, all_types


class Select(BaseFn):
    """Check if Select values are correct"""

    id = "E1017"
    shortdesc = "Select validation of parameters"
    description = "Making sure the Select function is properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-select.html"
    tags = ["functions", "select"]

    def __init__(self) -> None:
        super().__init__(
            "Fn::Select",
            all_types,
            resolved_rule="W1035",
        )
        self.fn_select = self.validate



class CfnLintExitException(Exception):
    """
    An exception that is raised to indicate that the CloudFormation linter should exit.

    This exception is used to signal that the linter should exit
    with a specific exit code, typically indicating the severity
    of the issues found in the CloudFormation template.

    Attributes:
        exit_code (int): The exit code to be used when the linter exits.

    Methods:
        __init__(self, exit_code: int) -> None:
            Initialize a new CfnLintExitException instance with the specified exit code.
    """

    def __init__(self, msg=None, exit_code=1):
        """
        Initialize a new CfnLintExitException instance with the specified exit code.

        Args:
            exit_code (int): The exit code to be used when the linter exits.
        """
        if msg is None:
            msg = f"process failed with exit code {exit_code}"
        super().__init__(msg)
        self.exit_code = exit_code


class InvalidRegionException(CfnLintExitException):
    """
    An exception that is raised when an invalid AWS region is encountered.

    This exception is raised when the CloudFormation linter encounters a resource
    or parameter that references an AWS region that is not valid or supported.
    """


class UnexpectedRuleException(CfnLintExitException):
    """
    An exception that is raised when an unexpected error occurs while loading rules.

    This exception is raised when the CloudFormation linter encounters an error
    while attempting to load custom rules or rules from a specified directory or
    module. This could be due to a variety of reasons, such as a missing file,
    a syntax error in the rule code, or an issue with the rule implementation.
    """


class DuplicateRuleError(CfnLintExitException):
    """
    The data associated with a particular path could not be loaded.
    :ivar data_path: The data path that the user attempted to load.
    """

    def __init__(self, rule_id: str):
        super().__init__(f"Rule already included: {rule_id}")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.other.functions
from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.rules.functions._BaseFn import BaseFn, SchemaDetails


class Equals(BaseFn):
    """Check Equals Condition Function Logic"""

    id = "E8003"
    shortdesc = "Check Fn::Equals structure for validity"
    description = "Check Fn::Equals is a list of two elements"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-conditions.html#intrinsic-function-reference-conditions-equals"
    tags = ["functions", "equals"]

    def __init__(self) -> None:
        super().__init__(
            "Fn::Equals",
            ("boolean",),
            schema_details=SchemaDetails(
                cfnlint.data.schemas.other.functions, "equals.json"
            ),
        )
        self.child_rules = {
            "W8003": None,
        }

    def fn_equals(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> ValidationResult:
        errs = list(self.validate(validator, s, instance, schema))
        if errs:
            yield from iter(errs)
            return

        child_rule = self.child_rules.get("W8003")
        if child_rule and hasattr(child_rule, "equals_is_useful"):
            yield from child_rule.equals_is_useful(
                validator, s, instance.get("Fn::Equals", []), schema
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema import CfnLintKeyword


class DependsOnObsolete(CfnLintKeyword):
    """Check unneeded DepensOn Resource Configuration"""

    id = "W3005"
    shortdesc = "Check obsolete DependsOn configuration for Resources"
    description = (
        "Check if DependsOn is specified if not needed. "
        "A Ref or a Fn::GetAtt already is an implicit dependency."
    )
    source_url = (
        "https://aws.amazon.com/blogs/devops/optimize-aws-cloudformation-templates/"
    )
    tags = ["resources", "dependson", "ref", "getatt"]

    def __init__(self) -> None:
        super().__init__(keywords=["Resources/*/DependsOn", "Resources/*/DependsOn/*"])

    def validate(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> ValidationResult:
        if not validator.is_type(instance, "string"):
            return

        from_resource_name = validator.context.path.path[1]
        if validator.cfn.graph is None:
            return
        edges = validator.cfn.graph.graph.get_edge_data(from_resource_name, instance)
        # returns None if no edge exists
        if not edges:
            return
        for _, edge in edges.items():
            if edge["label"] != "DependsOn":
                path = list(validator.context.path.path)[0:2] + edge["source_paths"]
                yield ValidationError(
                    f"{instance!r} dependency already enforced"
                    f" by a {edge['label']!r} at {'/'.join(str(e) for e in path)!r}",
                )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any

from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.rules.helpers import get_value_from_path
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class PipelineArtifactCounts(CfnLintKeyword):
    id = "E3702"
    shortdesc = "Validate the number of input and output artifacts in a CodePipeline"
    description = (
        "When using AWS::CodePipeline::Pipeline action types "
        "have different contraints for InputArtifacts and "
        "OutputArtifacts"
    )
    tags = ["resources", "codepipeline"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::CodePipeline::Pipeline/Properties/Stages/*/Actions/*",
            ],
        )

        self._artifact_counts: dict[str, dict[str, dict[str, Any]]] = {
            "AWS": {
                "Source": {
                    "S3": {
                        "InputArtifacts": {"maxItems": 0},
                        "OutputArtifacts": {"minItems": 1, "maxItems": 1},
                    },
                    "CodeCommit": {
                        "InputArtifacts": {"maxItems": 0},
                        "OutputArtifacts": {"minItems": 1, "maxItems": 1},
                    },
                    "ECR": {
                        "InputArtifacts": {"maxItems": 0},
                        "OutputArtifacts": {"minItems": 1, "maxItems": 1},
                    },
                },
                "Test": {
                    "CodeBuild": {
                        "InputArtifacts": {"minItems": 1, "maxItems": 5},
                        "OutputArtifacts": {"minItems": 0, "maxItems": 5},
                    },
                    "DeviceFarm": {
                        "InputArtifacts": {"minItems": 1, "maxItems": 1},
                        "OutputArtifacts": {"maxItems": 0},
                    },
                },
                "Build": {
                    "CodeBuild": {
                        "InputArtifacts": {"minItems": 1, "maxItems": 5},
                        "OutputArtifacts": {"minItems": 0, "maxItems": 5},
                    }
                },
                "Approval": {
                    "Manual": {
                        "InputArtifacts": {"maxItems": 0},
                        "OutputArtifacts": {"maxItems": 0},
                    }
                },
                "Deploy": {
                    "S3": {
                        "InputArtifacts": {"minItems": 1, "maxItems": 1},
                        "OutputArtifacts": {"maxItems": 0},
                    },
                    "CloudFormation": {
                        "InputArtifacts": {"minItems": 0, "maxItems": 10},
                        "OutputArtifacts": {"minItems": 0, "maxItems": 1},
                    },
                    "CodeDeploy": {
                        "InputArtifacts": {"minItems": 1, "maxItems": 1},
                        "OutputArtifacts": {"maxItems": 0},
                    },
                    "ElasticBeanstalk": {
                        "InputArtifacts": {"minItems": 1, "maxItems": 1},
                        "OutputArtifacts": {"maxItems": 0},
                    },
                    "OpsWorks": {
                        "InputArtifacts": {"minItems": 1, "maxItems": 1},
                        "OutputArtifacts": {"maxItems": 0},
                    },
                    "ECS": {
                        "InputArtifacts": {"minItems": 1, "maxItems": 1},
                        "OutputArtifacts": {"maxItems": 0},
                    },
                    "ServiceCatalog": {
                        "InputArtifacts": {"minItems": 1, "maxItems": 1},
                        "OutputArtifacts": {"maxItems": 0},
                    },
                },
                "Invoke": {
                    "Lambda": {
                        "InputArtifacts": {"minItems": 0, "maxItems": 5},
                        "OutputArtifacts": {"minItems": 0, "maxItems": 5},
                    }
                },
            },
            "ThirdParty": {
                "Source": {
                    "GitHub": {
                        "InputArtifacts": {"maxItems": 0},
                        "OutputArtifacts": {"minItems": 1, "maxItems": 1},
                    }
                },
                "Deploy": {
                    "AlexaSkillsKit": {
                        "InputArtifacts": {"minItems": 0, "maxItems": 2},
                        "OutputArtifacts": {"maxItems": 0},
                    },
                },
            },
            "Custom": {
                "Build": {
                    "Jenkins": {
                        "InputArtifacts": {"minItems": 0, "maxItems": 5},
                        "OutputArtifacts": {"minItems": 0, "maxItems": 5},
                    },
                },
                "Test": {
                    "Jenkins": {
                        "InputArtifacts": {"minItems": 0, "maxItems": 5},
                        "OutputArtifacts": {"minItems": 0, "maxItems": 5},
                    },
                },
            },
        }

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        if not validator.is_type(instance, "object"):
            return

        for action, action_validator in get_value_from_path(
            validator, instance, path=deque(["ActionTypeId"])
        ):
            for owner, owner_validator in get_value_from_path(
                action_validator, action, path=deque(["Owner"])
            ):
                owner_artifact_counts = self._artifact_counts.get(owner)
                if not owner_artifact_counts:
                    continue
                for category, categor_validator in get_value_from_path(
                    owner_validator, action, path=deque(["Category"])
                ):
                    category_artifact_counts = owner_artifact_counts.get(category)
                    if not category_artifact_counts:
                        continue
                    for provider, provider_provider in get_value_from_path(
                        categor_validator, action, deque(["Provider"])
                    ):
                        count_schema: dict[str, Any] = {
                            "properties": category_artifact_counts.get(provider),
                            "required": [],
                        }
                        if not count_schema.get("properties"):
                            continue

                        for key in ["InputArtifacts", "OutputArtifacts"]:
                            if (
                                count_schema.get("properties", {})
                                .get(key, {})
                                .get("minItems", 0)
                                > 0
                            ):
                                count_schema["required"].append(key)

                        provider_provider = provider_provider.evolve(
                            schema=count_schema,
                        )

                        repr = {
                            "Owner": owner,
                            "Category": category,
                            "Provider": provider,
                        }
                        for err in provider_provider.iter_errors(instance):
                            err.message = f"{err.message} when using {repr!r}"
                            err.rule = self
                            yield err



__all__ = [
    "ResourceNotFoundError",
    "AttributeDict",
    "GetAtt",
    "GetAttType",
    "PROVIDER_SCHEMA_MANAGER",
    "OTHER_SCHEMA_MANAGER",
    "Schema",
    "SchemaPatch",
    "patch",
    "reset",
]

from cfnlint.schema._exceptions import ResourceNotFoundError
from cfnlint.schema._getatts import AttributeDict
from cfnlint.schema._patch import SchemaPatch, patch, reset
from cfnlint.schema._schema import Schema
from cfnlint.schema.manager import PROVIDER_SCHEMA_MANAGER
from cfnlint.schema.other_schema_manager import OTHER_SCHEMA_MANAGER



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules.errors.config import ConfigError
from cfnlint.rules.errors.parse import ParseError
from cfnlint.rules.errors.rule import RuleError
from cfnlint.rules.errors.transform import TransformError



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.rules.Configuration import Configuration


@pytest.fixture(scope="module")
def rule():
    rule = Configuration()
    yield rule


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Empty is okay",
            {},
            [],
        ),
        (
            "wrong type",
            [],
            [
                ValidationError(
                    "[] is not of type 'object'",
                    validator="type",
                    schema_path=deque(["type"]),
                    rule=Configuration(),
                )
            ],
        ),
        (
            "Wrong type of rule",
            {
                "Rule1": [],
            },
            [
                ValidationError(
                    "[] is not of type 'object'",
                    validator="type",
                    schema_path=deque(["patternProperties", "^[A-Za-z0-9]+$", "type"]),
                    path=deque(["Rule1"]),
                    rule=Configuration(),
                )
            ],
        ),
        (
            "Empty rule",
            {
                "Rule1": {},
            },
            [
                ValidationError(
                    "'Assertions' is a required property",
                    validator="required",
                    schema_path=deque(
                        ["patternProperties", "^[A-Za-z0-9]+$", "required"]
                    ),
                    path=deque(["Rule1"]),
                    rule=Configuration(),
                )
            ],
        ),
        (
            "Valid rule with RuleCondition and Assertions",
            {
                "Rule1": {
                    "RuleCondition": {"Fn::Equals": ["a", "b"]},
                    "Assertions": [
                        {
                            "Assert": {"Fn::Equals": ["a", "b"]},
                            "AssertDescription": "a is equal to b",
                        },
                        {
                            "Assert": {"Fn::Equals": ["a", "b"]},
                        },
                    ],
                },
            },
            [],
        ),
    ],
)
def test_validate(name, instance, expected, rule, validator):
    errs = list(rule.validate(validator, {}, instance, {}))

    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.ectwo.InstanceImageId import InstanceImageId


@pytest.fixture(scope="module")
def rule():
    rule = InstanceImageId()
    yield rule


@pytest.mark.parametrize(
    "name,template,instance,path,expected",
    [
        (
            "Valid with no ImageId in LaunchTemplate",
            {
                "Resources": {
                    "LaunchTemplate": {
                        "Type": "AWS::EC2::LaunchTemplate",
                        "Properties": {
                            "LaunchTemplateName": "a-template",
                            "LaunchTemplateData": {
                                "Monitoring": {"Enabled": True},
                            },
                        },
                    },
                }
            },
            {
                "LaunchTemplate": {
                    "LaunchTemplateId": {
                        "Fn::GetAtt": [
                            "LaunchTemplate",
                            "LaunchTemplateId",
                        ],
                    }
                },
                "ImageId": "ami-12345678",
            },
            {
                "path": ["Resources", "Instance", "Properties"],
            },
            [],
        ),
        (
            "Valid with no ImageId and a string launch template",
            {
                "Resources": {},
            },
            {
                "LaunchTemplate": {"LaunchTemplateId": "foo"},
            },
            {
                "path": ["Resources", "Instance", "Properties"],
            },
            [],
        ),
        (
            "Valid with no ImageId and a parameter",
            {
                "Parameters": {
                    "LaunchTemplateId": {
                        "Type": "String",
                    }
                },
                "Resources": {},
            },
            {
                "LaunchTemplate": {"LaunchTemplateId": {"Ref": "LaunchTemplateId"}},
            },
            {
                "path": ["Resources", "Instance", "Properties"],
            },
            [],
        ),
        (
            "Valid with no ImageId a random ref",
            {
                "Resources": {},
            },
            {
                "LaunchTemplate": {"LaunchTemplateName": {"Ref": "AWS::StackName"}},
            },
            {
                "path": ["Resources", "Instance", "Properties"],
            },
            [],
        ),
        (
            "Valid with no ImageId and another function",
            {
                "Parameters": {
                    "LaunchTemplateId": {
                        "Type": "String",
                    }
                },
                "Resources": {},
            },
            {
                "LaunchTemplate": {
                    "LaunchTemplateId": {"Fn::FindInMap": ["One", "Two", "Three"]}
                },
            },
            {
                "path": ["Resources", "Instance", "Properties"],
            },
            [],
        ),
        (
            "Valid with ImageId in LaunchTemplate",
            {
                "Resources": {
                    "LaunchTemplate": {
                        "Type": "AWS::EC2::LaunchTemplate",
                        "Properties": {
                            "LaunchTemplateName": "a-template",
                            "LaunchTemplateData": {
                                "Monitoring": {"Enabled": True},
                                "ImageId": "ami-12345678",
                            },
                        },
                    },
                }
            },
            {
                "LaunchTemplate": {
                    "LaunchTemplateId": {
                        "Fn::GetAtt": [
                            "LaunchTemplate",
                            "LaunchTemplateId",
                        ],
                    }
                },
            },
            {
                "path": ["Resources", "Instance", "Properties"],
            },
            [],
        ),
        (
            "Valid with ImageId in LaunchTemplate using Name",
            {
                "Resources": {
                    "LaunchTemplate": {
                        "Type": "AWS::EC2::LaunchTemplate",
                        "Properties": {
                            "LaunchTemplateName": "a-template",
                            "LaunchTemplateData": {
                                "Monitoring": {"Enabled": True},
                                "ImageId": "ami-12345678",
                            },
                        },
                    },
                }
            },
            {
                "LaunchTemplate": {
                    "LaunchTemplateName": {
                        "Ref": "LaunchTemplate",
                    }
                },
            },
            {
                "path": ["Resources", "Instance", "Properties"],
            },
            [],
        ),
        (
            "Invalid with no relationship",
            {},
            {},
            {
                "path": ["Resources", "Instance", "Properties"],
            },
            [
                ValidationError(
                    "'ImageId' is a required property",
                    validator="required",
                    rule=InstanceImageId(),
                    path=deque([]),
                )
            ],
        ),
        (
            "Invalid with no ImageId in LaunchTemplate",
            {
                "Resources": {
                    "LaunchTemplate": {
                        "Type": "AWS::EC2::LaunchTemplate",
                        "Properties": {
                            "LaunchTemplateName": "a-template",
                            "LaunchTemplateData": {
                                "Monitoring": {"Enabled": True},
                            },
                        },
                    },
                }
            },
            {
                "LaunchTemplate": {
                    "LaunchTemplateId": {
                        "Fn::GetAtt": [
                            "LaunchTemplate",
                            "LaunchTemplateId",
                        ],
                    }
                },
            },
            {
                "path": ["Resources", "Instance", "Properties"],
            },
            [
                ValidationError(
                    "'ImageId' is a required property",
                    validator="required",
                    rule=InstanceImageId(),
                    path=deque([]),
                )
            ],
        ),
        (
            "Invalid with no ImageId in LaunchTemplate with condition",
            {
                "Conditions": {
                    "IsUsEast1": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]}
                },
                "Resources": {
                    "LaunchTemplate": {
                        "Type": "AWS::EC2::LaunchTemplate",
                        "Properties": {
                            "LaunchTemplateName": "a-template",
                            "LaunchTemplateData": {
                                "Monitoring": {"Enabled": True},
                            },
                        },
                    },
                },
            },
            {
                "LaunchTemplate": {
                    "LaunchTemplateId": {
                        "Fn::GetAtt": [
                            "LaunchTemplate",
                            "LaunchTemplateId",
                        ],
                    }
                },
                "ImageId": {
                    "Fn::If": [
                        "IsUsEast1",
                        "ami-12345678",
                        {"Ref": "AWS::NoValue"},
                    ]
                },
            },
            {
                "path": ["Resources", "Instance", "Properties"],
            },
            [
                ValidationError(
                    "'ImageId' is a required property",
                    validator="required",
                    rule=InstanceImageId(),
                    path=deque(["ImageId", "Fn::If", 2, "Ref"]),
                )
            ],
        ),
    ],
    indirect=["template", "path"],
)
def test_validate(name, instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert (
        errs == expected
    ), f"Expected test {name!r} to have {expected!r} but got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.other.resources
from cfnlint.helpers import FUNCTIONS
from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class CfnInit(CfnLintJsonSchema):
    """Check CloudFormation Init items"""

    id = "E3009"
    shortdesc = "Check CloudFormation init configuration"
    description = "Validate that the items in a CloudFormation init adhere to standards"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-init.html#aws-resource-cloudformation-init-syntax"
    tags = ["resources", "cloudformation init"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/*/Metadata/AWS::CloudFormation::Init"],
            schema_details=SchemaDetails(
                cfnlint.data.schemas.other.resources, "cfn_init.json"
            ),
            all_matches=True,
        )

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        cfn_validator = self.extend_validator(
            validator=validator,
            schema=self._schema,
            context=validator.context.evolve(functions=list(FUNCTIONS)),
        )

        yield from super()._iter_errors(cfn_validator, instance)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.helpers import bool_compare
from cfnlint.rules import CloudFormationLintRule, RuleMatch


class CacheClusterFailover(CloudFormationLintRule):
    """Check automatic failover on a cache cluster"""

    id = "E3026"
    shortdesc = "Check Elastic Cache Redis Cluster settings"
    description = (
        "Evaluate Redis Cluster groups to make sure automatic failover is "
        "enabled when cluster mode is enabled"
    )
    source_url = "https://github.com/aws-cloudformation/cfn-lint"
    tags = ["resources", "elasticcache"]

    def __init__(self):
        """Init"""
        super().__init__()
        self.resource_property_types.append("AWS::ElastiCache::ReplicationGroup")

    def is_cluster_enabled(self, properties):
        """Test if cluster is enabled"""
        if isinstance(properties, dict):
            for property_name, property_value in properties.items():
                if property_name == "cluster-enabled" and property_value == "yes":
                    return True

        return False

    def _test_cluster_settings(
        self, properties, path, pg_properties, pg_path, cfn, scenario
    ):
        """test for each scenario"""
        results = []
        pg_conditions = cfn.get_conditions_from_path(cfn.template, pg_path)
        # test to make sure that any condition that may apply to the path for the Ref
        # is not applicable
        if pg_conditions and scenario:
            for c_name, c_value in scenario.items():
                if c_name in pg_conditions:
                    if c_value not in pg_conditions.get(c_name):
                        return results
        if self.is_cluster_enabled(
            cfn.get_value_from_scenario(pg_properties, scenario)
        ):
            c_props = cfn.get_value_from_scenario(properties, scenario)
            automatic_failover = c_props.get("AutomaticFailoverEnabled")
            if bool_compare(automatic_failover, False):
                pathmessage = path[:] + ["AutomaticFailoverEnabled"]
                if scenario is None:
                    message = (
                        '"AutomaticFailoverEnabled" must be misssing or True when'
                        " setting up a cluster at {0}"
                    )
                    results.append(
                        RuleMatch(
                            pathmessage, message.format("/".join(map(str, pathmessage)))
                        )
                    )
                else:
                    message = (
                        '"AutomaticFailoverEnabled" must be misssing or True when'
                        " setting up a cluster when {0} at {1}"
                    )
                    scenario_text = " and ".join(
                        [f'when condition "{k}" is {v}' for (k, v) in scenario.items()]
                    )
                    results.append(
                        RuleMatch(
                            pathmessage,
                            message.format(
                                scenario_text, "/".join(map(str, pathmessage))
                            ),
                        )
                    )
            num_node_groups = c_props.get("NumNodeGroups")
            if not num_node_groups:
                # only test cache nodes if num node groups aren't specified
                num_cache_nodes = c_props.get("NumCacheClusters", 0)
                if num_cache_nodes <= 1:
                    pathmessage = path[:] + ["NumCacheClusters"]
                    if scenario is None:
                        message = (
                            '"NumCacheClusters" must be greater than one when creating'
                            " a cluster at {0}"
                        )
                        results.append(
                            RuleMatch(
                                pathmessage,
                                message.format("/".join(map(str, pathmessage))),
                            )
                        )
                    else:
                        message = (
                            '"NumCacheClusters" must be greater than one when creating'
                            " a cluster when {0} at {1}"
                        )
                        scenario_text = " and ".join(
                            [
                                f'when condition "{k}" is {v}'
                                for (k, v) in scenario.items()
                            ]
                        )
                        results.append(
                            RuleMatch(
                                pathmessage,
                                message.format(
                                    scenario_text, "/".join(map(str, pathmessage))
                                ),
                            )
                        )

        return results

    def test_cluster_settings(self, properties, path, pg_resource_name, pg_path, cfn):
        """Test cluster settings for the parameter group and Replication Group"""
        results = []
        pg_properties = (
            cfn.template.get("Resources", {})
            .get(pg_resource_name, {})
            .get("Properties", {})
            .get("Properties", {})
        )
        scenarios = cfn.get_conditions_scenarios_from_object(
            [properties, pg_properties]
        )
        if scenarios:
            for scenario in scenarios:
                results.extend(
                    self._test_cluster_settings(
                        properties, path, pg_properties, pg_path, cfn, scenario
                    )
                )
        else:
            results.extend(
                self._test_cluster_settings(
                    properties, path, pg_properties, pg_path, cfn, None
                )
            )
        return results

    def _check_ref(self, value, path, **kwargs):
        cfn = kwargs["cfn"]
        properties = kwargs["properties"]
        if value in cfn.get_resources():
            return self.test_cluster_settings(properties, path, value, path, cfn)
        return []

    def match_resource_properties(self, properties, _, path, cfn):
        """Check CloudFormation Properties"""

        return cfn.check_value(
            properties,
            "CacheParameterGroupName",
            path,
            check_ref=self._check_ref,
            properties=properties,
            cfn=cfn,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

import cfnlint.data.schemas.other.conditions
from cfnlint.helpers import FUNCTION_CONDITIONS
from cfnlint.jsonschema import Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails
from cfnlint.schema.resolver import RefResolver


class Configuration(CfnLintJsonSchema):
    """Check if Conditions are configured correctly"""

    id = "E8001"
    shortdesc = "Conditions have appropriate properties"
    description = "Check if Conditions are properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/conditions-section-structure.html"
    tags = ["conditions"]

    def __init__(self):
        super().__init__(
            keywords=["Conditions"],
            schema_details=SchemaDetails(
                cfnlint.data.schemas.other.conditions,
                "conditions.json",
            ),
            all_matches=True,
        )

    def validate(
        self, validator: Validator, conditions: Any, instance: Any, schema: Any
    ):
        validator = validator.evolve(
            context=validator.context.evolve(
                functions=list(FUNCTION_CONDITIONS) + ["Condition"],
                resources={},
                strict_types=False,
            ),
            schema=self._schema,
            resolver=RefResolver.from_schema(
                self._schema,
            ),
        )

        yield from self._iter_errors(validator, instance)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import cfnlint.data.schemas.other.metadata
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class InterfaceConfiguration(CfnLintJsonSchema):
    """Check if Metadata Interface Configuration are configured correctly"""

    id = "E4001"
    shortdesc = "Metadata Interface have appropriate properties"
    description = "Metadata Interface properties are properly configured"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-cloudformation-interface.html"
    tags = ["metadata"]

    def __init__(self) -> None:
        super().__init__(
            keywords=["Metadata/AWS::CloudFormation::Interface"],
            schema_details=SchemaDetails(
                cfnlint.data.schemas.other.metadata, "interface.json"
            ),
            all_matches=True,
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import cfnlint.data.schemas.other.functions
from cfnlint.rules.functions._BaseFn import BaseFn, SchemaDetails


class Not(BaseFn):
    """Check Not Condition Function Logic"""

    id = "E8005"
    shortdesc = "Check Fn::Not structure for validity"
    description = "Check Fn::Not is a list of one element"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-conditions.html#intrinsic-function-reference-conditions-not"
    tags = ["functions", "not"]

    def __init__(self) -> None:
        super().__init__(
            "Fn::Not",
            ("boolean",),
            schema_details=SchemaDetails(
                cfnlint.data.schemas.other.functions, "not.json"
            ),
        )
        self.fn_not = self.validate

    # if validator.context.path.path and validator.context.path.path[0] == "Rules":
    #        functions = list(FUNCTION_RULES)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import logging
from test.testlib.testcase import BaseTestCase

from cfnlint.schema.resolver import RefResolver
from cfnlint.schema.resolver._exceptions import RefResolutionError

LOGGER = logging.getLogger("cfnlint.schema.manager")
LOGGER.disabled = True


class TestPointer(BaseTestCase):
    """Used for Pointer"""

    def setUp(self) -> None:
        super().setUp()
        self.schema = {
            "typeName": "Test::Resource::Type",
            "definitions": {
                "fooBar": {
                    "type": "object",
                    "properties": {"a": {"type": "string"}, "b": {"type": "string"}},
                },
                "anyOf": {
                    "anyOf": [
                        {"$ref": "#/definitions/fooBar"},
                        {
                            "type": "object",
                            "properties": {
                                "a": {"type": "boolean"},
                                "c": {"type": "boolean"},
                                "d": {"type": "boolean"},
                            },
                        },
                    ]
                },
            },
            "properties": {
                "foo": {"type": "string"},
                "bar": {"$ref": "#/definitions/bar"},
                "fooBars": {"type": "array", "items": {"$ref": "#/definitions/fooBar"}},
                "refFirst": {
                    "$ref": "#/definitions/fooBar",
                    "type": "object",
                    "properties": {
                        "a": {"type": "boolean"},
                    },
                },
                "anyOf": {"$ref": "#/definitions/anyOf"},
            },
        }

    def test_pointer(self):
        resolver = RefResolver(self.schema)
        self.assertEqual(
            resolver.resolve_cfn_pointer("/properties/foo"),
            {"type": "string"},
        )
        self.assertEqual(
            resolver.resolve_cfn_pointer("/properties/fooBars/*/a"),
            {"type": "string"},
        )

        # first one found is string (not boolean)
        self.assertEqual(
            resolver.resolve_cfn_pointer("/properties/anyOf/a"),
            {"type": "string"},
        )

        # second option in anyOf has the property we are looking for
        self.assertEqual(
            resolver.resolve_cfn_pointer("/properties/anyOf/c"),
            {"type": "boolean"},
        )

        # refs are handled first
        self.assertEqual(
            resolver.resolve_cfn_pointer("/properties/refFirst/a"),
            {"type": "string"},
        )

        with self.assertRaises(RefResolutionError):
            resolver.resolve_cfn_pointer("/properties/bar/key")



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.config import ConfigMixIn
from cfnlint.runner import Runner


@pytest.mark.parametrize(
    "name,config,expected_count",
    [
        (
            "Test decode errors with multiple files",
            ConfigMixIn(
                cli_args=[
                    "--template",
                    "test/fixtures/templates/bad/duplicate.yaml",
                    "test/fixtures/templates/bad/duplicate.json",
                ]
            ),
            6,
        ),
        (
            "Test decode errors with E0000 being ignored",
            ConfigMixIn(
                cli_args=[
                    "--template",
                    "test/fixtures/templates/bad/core/parse_invalid_map.yaml",
                    "--ignore-bad-template",
                ]
            ),
            0,
        ),
        (
            "Test decode return E0000 errors",
            ConfigMixIn(
                cli_args=[
                    "--template",
                    "test/fixtures/templates/bad/core/parse_invalid_map.yaml",
                ]
            ),
            1,
        ),
    ],
)
def test_run(name, config, expected_count):

    runner = Runner(config)

    errs = list(runner.run())

    assert len(errs) == expected_count, f"{name}: {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import Path, create_context_for_template
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.parameters.DynamicReferenceSecret import DynamicReferenceSecret
from cfnlint.template import Template


@pytest.fixture(scope="module")
def rule():
    rule = DynamicReferenceSecret()
    yield rule


@pytest.fixture(scope="module")
def cfn():
    return Template(
        "",
        {
            "Parameters": {
                "MyParameter": {
                    "Type": "String",
                }
            },
            "Resources": {},
        },
        regions=["us-east-1"],
    )


@pytest.fixture(scope="module")
def context(cfn):
    return create_context_for_template(cfn)


@pytest.mark.parametrize(
    "name,instance,path,expected",
    [
        (
            "REFing a parameter without a string",
            [],
            deque([]),
            [],
        ),
        (
            "REFing a resource=",
            "MyResource",
            deque([]),
            [],
        ),
        (
            "REFing a parameter",
            "MyParameter",
            deque([]),
            [
                ValidationError(
                    "Use dynamic references over parameters for secrets",
                    rule=DynamicReferenceSecret(),
                )
            ],
        ),
        (
            "REFing a parameter in a sub",
            "MyParameter",
            deque(["Fn::Sub"]),
            [],
        ),
    ],
)
def test_validate(name, instance, path, expected, rule, context, cfn):
    validator = CfnTemplateValidator(context=context.evolve(path=Path(path)), cfn=cfn)
    errs = list(rule.validate(validator, {}, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from copy import copy
from dataclasses import dataclass, field
from typing import Any, Dict, Set

from cfnlint._typing import RuleMatches
from cfnlint.jsonschema._utils import equal
from cfnlint.rules import CloudFormationLintRule, RuleMatch
from cfnlint.schema import PROVIDER_SCHEMA_MANAGER, ResourceNotFoundError
from cfnlint.template import Template


@dataclass
class _Seen:
    conditions: Dict[str, Set[bool]] = field(init=True, default_factory=dict)
    resources: Set[str] = field(init=True, default_factory=set)
    instance: Dict[str, Any] = field(init=True, default_factory=dict)


class PrimaryIdentifiers(CloudFormationLintRule):
    id = "E3019"
    shortdesc = "Validate that all resources have unique primary identifiers"
    description = (
        "Use the primary identifiers in a resource schema to validate that "
        "resources inside the template are unique"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/resources-section-structure.html"
    tags = ["parameters", "resources"]

    def __init__(self):
        super().__init__()

        self.config_definition = {
            "exceptions": {
                "default": {
                    "AWS::CodeBuild::Project": {
                        "primaryIdentifier": ["/properties/Name"]
                    }
                },
                "type": "object",
                "itemtype": "object",
            }
        }
        self.configure()

    def _merge_conditions(
        self, conditions1: Dict[str, Set[bool]], conditions2: Dict[str, Set[bool]]
    ):
        for k2, v2 in conditions2.items():
            if k2 in conditions1:
                if v2 != conditions1[k2]:
                    raise ValueError("Condition mismatch")
            conditions1[k2] = v2

        return conditions1

    def _validate_resource_type_uniqueness(self, cfn, resource_type, ids):
        matches = []
        seens = []
        for resource_name, resource_attributes in cfn.get_resources(
            resource_type
        ).items():
            condition = resource_attributes.get("Condition")
            properties = resource_attributes.get("Properties", {})
            path = ["Resources", resource_name, "Properties"]
            for scenario in cfn.get_object_without_conditions(properties):
                object = scenario.get("Object")
                conditions = scenario.get("Scenario")
                if conditions is None:
                    conditions = {}
                # we need to change this to a set for implies to work
                conditions = {k: set([v]) for k, v in conditions.items()}

                # if there is a resource level condition lets see if its valid
                # and add it to the scenario as needed
                if condition:
                    if scenario:
                        if not cfn.conditions.check_implies(scenario, condition):
                            continue
                    conditions[condition] = set([True])

                # validate object is an object just in case it was emptied out
                # from the condition (Ref: AWS::NoValue)
                if not isinstance(object, dict):
                    continue

                # get_object_without_conditions will return all elements
                # so we are going to filter down to the ids that we care about
                identifiers = {id: object.get(id) for id in ids}
                # if any of the identifiers are None that means the service
                # will define it and we can skip this test
                if any(identifier is None for identifier in identifiers.values()):
                    continue

                # build the element
                element = _Seen(conditions, set([resource_name]), identifiers)

                found = False
                for seen in seens:
                    if equal(seen.instance, element.instance):
                        try:
                            merged_conditions = self._merge_conditions(
                                copy(seen.conditions), copy(element.conditions)
                            )
                            # if conditions are empty we found it
                            if not merged_conditions:
                                found = True
                                seen.resources.add(resource_name)
                            for _ in cfn.conditions.build_scenarios(merged_conditions):
                                seen.resources.add(resource_name)
                                # because the conditions could be unique
                                # we will validate conditions being equal
                                # and only say found if they match
                                if equal(element.conditions, seen.conditions):
                                    found = True
                                break

                        except ValueError:
                            continue

                # add anything not already Found to the seens element
                if not found:
                    seens.append(element)

        for seen in seens:
            # build errors for any seen items with more than 1 resource
            if len(seen.resources) > 1:
                for resource in seen.resources:
                    path = ["Resources", resource, "Properties"]
                    if len(seen.instance) == 1:
                        key = list(seen.instance.keys())[0]
                        path.append(key)

                    message = (
                        f"Primary identifiers {seen.instance!r} should "
                        "have unique values across the resources "
                        f"{seen.resources!r}"
                    )

                    matches.append(RuleMatch(path, message))

        return matches

    def match(self, cfn: Template) -> RuleMatches:
        tS = set()
        for _, resource_properties in cfn.get_resources().items():
            t = resource_properties.get("Type")
            if isinstance(t, str):
                tS.add(t)

        matches = []
        for t in tS:
            try:
                if t in self.config.get("exceptions", {}):
                    schema = self.config.get("exceptions", {}).get(t)
                else:
                    schema = PROVIDER_SCHEMA_MANAGER.get_resource_schema(
                        cfn.regions[0], t
                    ).schema

                # we are worried about primary identifiers that can be set
                # by the customer so if any primary identifiers are read
                # only we have to skip evaluation
                primary_ids = schema.get("primaryIdentifier", [])
                if not primary_ids:
                    continue

                read_only_ids = schema.get("readOnlyProperties", [])

                if any(id in read_only_ids for id in primary_ids):
                    continue

                primary_ids = [id.replace("/properties/", "") for id in primary_ids]
                # most primaryIdentifiers are at the root level. At this time
                # we can only validate if they are at the root of Properties
                if any("/" in id for id in primary_ids):
                    continue
                matches.extend(
                    self._validate_resource_type_uniqueness(cfn, t, primary_ids)
                )
            except ResourceNotFoundError:
                continue
        return matches



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import logging
from pathlib import Path
from test.testlib.testcase import BaseTestCase
from unittest.mock import patch

import cfnlint.config
from cfnlint.jsonschema import ValidationError

LOGGER = logging.getLogger("cfnlint")


class TestConfigFileArgs(BaseTestCase):
    """Test ConfigParser Arguments"""

    def tearDown(self):
        """Setup"""
        for handler in LOGGER.handlers:
            LOGGER.removeHandler(handler)

    def test_config_parser_read_config(self):
        """Testing one file successful"""
        config = cfnlint.config.ConfigFileArgs(
            config_file=Path("test/fixtures/configs/cfnlintrc_read.yaml")
        )
        self.assertEqual(
            config.file_args,
            {
                "templates": ["test/fixtures/templates/good/**/*.yaml"],
                "include_checks": ["I"],
            },
        )

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_parser_read_config_only_one_read(self, yaml_mock):
        """Testing one file successful"""
        # Should only have one read here
        # We aren't going to search and find the possible locations of
        # cfnlintrc
        yaml_mock.side_effect = [
            {"regions": ["us-west-1"]},
        ]
        config = cfnlint.config.ConfigFileArgs(
            config_file=Path("test/fixtures/configs/cfnlintrc_read.yaml")
        )
        self.assertEqual(config.file_args, {"regions": ["us-west-1"]})

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_parser_read(self, yaml_mock):
        """Testing one file successful"""
        yaml_mock.side_effect = [{"regions": ["us-west-1"]}, {}]
        results = cfnlint.config.ConfigFileArgs()
        self.assertEqual(results.file_args, {"regions": ["us-west-1"]})

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_parser_read_merge(self, yaml_mock):
        """test the merge of config"""

        yaml_mock.side_effect = [{"regions": ["us-west-1"]}, {"regions": ["us-east-1"]}]

        results = cfnlint.config.ConfigFileArgs()
        self.assertEqual(results.file_args, {"regions": ["us-east-1"]})

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_parser_fail_on_bad_config(self, yaml_mock):
        """test the read call to the config parser is reading two files"""

        yaml_mock.side_effect = [{"regions": True}, {}]

        with self.assertRaises(ValidationError):
            cfnlint.config.ConfigFileArgs()

    @patch("cfnlint.config.ConfigFileArgs._read_config", create=True)
    def test_config_parser_fail_on_config_rules(self, yaml_mock):
        """
        test the read call to the config parser is
        parsing configure rules correctly
        """

        yaml_mock.side_effect = [{"configure_rules": {"E3012": {"strict": False}}}, {}]

        results = cfnlint.config.ConfigFileArgs()
        self.assertEqual(
            results.file_args, {"configure_rules": {"E3012": {"strict": False}}}
        )

    @patch("pathlib.Path.is_file", create=True)
    def test_config_parser_is_file_both(self, is_file_mock):
        calls = [
            True,
            True,
            False,
            False,
        ]
        is_file_mock.side_effect = calls
        my_config = cfnlint.config.ConfigFileArgs()
        self.assertEqual(my_config._user_config_file.name, ".cfnlintrc")
        self.assertEqual(my_config._project_config_file.name, ".cfnlintrc")
        self.assertEqual(is_file_mock.call_count, len(calls))

    @patch("pathlib.Path.is_file", create=True)
    def test_config_parser_is_file_both_yaml(self, is_file_mock):
        calls = [
            False,
            True,
            False,
            True,
            False,
            False,
        ]
        is_file_mock.side_effect = calls
        my_config = cfnlint.config.ConfigFileArgs()
        self.assertEqual(my_config._user_config_file.name, ".cfnlintrc.yaml")
        self.assertEqual(my_config._project_config_file.name, ".cfnlintrc.yaml")
        self.assertEqual(is_file_mock.call_count, len(calls))

    @patch("pathlib.Path.is_file", create=True)
    def test_config_parser_is_file_both_yml(self, is_file_mock):
        calls = [
            False,
            False,
            True,
            False,
            False,
            True,
            False,
            False,
        ]
        is_file_mock.side_effect = calls
        my_config = cfnlint.config.ConfigFileArgs()
        self.assertEqual(my_config._user_config_file.name, ".cfnlintrc.yml")
        self.assertEqual(my_config._project_config_file.name, ".cfnlintrc.yml")
        self.assertEqual(is_file_mock.call_count, len(calls))



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.properties.Tagging import Tagging


@pytest.fixture(scope="module")
def rule():
    rule = Tagging()
    yield rule


@pytest.mark.parametrize(
    "name,instance,schema,expected",
    [
        (
            "Valid tag by array",
            {"Tags": [{"Key": "Foo", "Value": "Bar"}]},
            {"taggable": True},
            [],
        ),
        (
            "Valid tag by key/value",
            {"Tags": {"Foo": "Bar"}},
            {"taggable": True},
            [],
        ),
        (
            "Duplicate key",
            {"Tags": [{"Key": "Foo", "Value": "Bar"}, {"Key": "Foo", "Value": "Bar"}]},
            {"taggable": True},
            [
                ValidationError(
                    ("array items are not unique for keys ['Key']"),
                    path=deque(["Tags"]),
                    schema_path=deque(["properties", "Tags", "uniqueKeys"]),
                    validator="tagging",
                )
            ],
        ),
        (
            "Value is too long",
            {
                "Tags": [
                    {"Key": "Foo", "Value": "a" * 257},
                ]
            },
            {"taggable": True},
            [
                ValidationError(
                    "expected maximum length: 256, found: 257",
                    path=deque(["Tags", 0, "Value"]),
                    schema_path=deque(
                        [
                            "properties",
                            "Tags",
                            "items",
                            "properties",
                            "Value",
                            "maxLength",
                        ]
                    ),
                    validator="tagging",
                ),
            ],
        ),
        (
            "Value is too long in object",
            {
                "Tags": {"Foo": "a" * 257},
            },
            {"taggable": True},
            [
                ValidationError(
                    "expected maximum length: 256, found: 257",
                    path=deque(["Tags", "Foo"]),
                    schema_path=deque(
                        [
                            "properties",
                            "Tags",
                            "patternProperties",
                            "^(?!aws:).+$",
                            "maxLength",
                        ]
                    ),
                    validator="tagging",
                ),
            ],
        ),
        (
            "AWS key name in array",
            {
                "Tags": [
                    {"Key": "aws:Foo", "Value": "Bar"},
                ],
            },
            {"taggable": True},
            [
                ValidationError(
                    ("'aws:Foo' does not match " "'^(?!aws:).+$'"),
                    path=deque(["Tags", 0, "Key"]),
                    schema_path=deque(
                        ["properties", "Tags", "items", "properties", "Key", "pattern"]
                    ),
                    validator="tagging",
                ),
            ],
        ),
        (
            "AWS key name in object",
            {
                "Tags": {"aws:Foo": "Bar"},
            },
            {"taggable": True},
            [
                ValidationError(
                    (
                        "'aws:Foo' does not match any of "
                        "the regexes: "
                        "'^(?!aws:).+$'"
                    ),
                    path=deque(["Tags", "aws:Foo"]),
                    schema_path=deque(["properties", "Tags", "additionalProperties"]),
                    validator="tagging",
                ),
            ],
        ),
    ],
)
def test_validate(name, instance, schema, expected, rule, validator):
    errs = list(rule.tagging(validator, schema, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import datetime
import json
from typing import Any

import regex as re

from cfnlint.helpers import FUNCTIONS, ensure_list, is_function
from cfnlint.jsonschema import ValidationError
from cfnlint.rules import CloudFormationLintRule


class StringLength(CloudFormationLintRule):
    """Check if a String has a length within the limit"""

    id = "E3033"
    shortdesc = "Check if a string has between min and max number of values specified"
    description = "Check strings for its length between the minimum and maximum"
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/cfn-schema-specification.md#length"
    tags = ["resources", "property", "string", "size"]

    def _serialize_date(self, obj):
        if isinstance(obj, datetime.date):
            return obj.isoformat()
        return json.JSONEncoder.default(self, o=obj)

    def _fix_sub_string(self, instance):
        return re.sub(r"\${[a-zA-Z0-9._-]{1,255}}", "", instance)

    # pylint: disable=too-many-return-statements
    def _remove_functions(self, obj: Any) -> Any:
        """Replaces intrinsic functions with string"""
        if isinstance(obj, dict):
            new_obj = {}
            if len(obj) == 1:
                k = next(iter(obj))
                v = obj[k]
                if k in FUNCTIONS:
                    if k == "Fn::Sub":
                        if isinstance(v, str):
                            return self._fix_sub_string(v)
                        if isinstance(v, list):
                            return self._fix_sub_string(v[0])
                    else:
                        return ""
                else:
                    new_obj[k] = self._remove_functions(v)
                    return new_obj
            else:
                for k, v in obj.items():
                    new_obj[k] = self._remove_functions(v)
                return new_obj
        elif isinstance(obj, list):
            new_list = []
            for v in obj:
                new_list.append(self._remove_functions(v))
            return new_list

        return obj

    def _non_string_max_length(self, instance, mL):
        j = self._remove_functions(instance)
        if len(json.dumps(j, separators=(",", ":"), default=self._serialize_date)) > mL:
            yield ValidationError("Item is too long")

    def _non_string_min_length(self, instance, mL):
        j = self._remove_functions(instance)
        if len(json.dumps(j, separators=(",", ":"), default=self._serialize_date)) < mL:
            yield ValidationError("Item is too short")

    # pylint: disable=unused-argument, arguments-renamed
    def maxLength(self, validator, mL, instance, schema):
        if validator.is_type(instance, "string"):
            if schema.get("format") == "json":
                try:
                    instance = json.loads(instance)
                except:  # noqa: E722
                    pass
                    return
                yield from self._non_string_max_length(instance, mL)
                return
            else:
                if len(instance) > mL:
                    yield ValidationError(f"{instance!r} is longer than {mL}")
                return

        # there are scenarios where Fn::Sub may not predictable so use
        # best judgement
        key, value = is_function(instance)
        if key is not None:
            if key == "Fn::Sub":
                value = instance[key]
                if isinstance(value, str):
                    yield from self.maxLength(
                        validator, mL, self._fix_sub_string(value), schema
                    )
                elif isinstance(value, list) and len(value) == 2:
                    yield from self.maxLength(
                        validator, mL, self._fix_sub_string(value[0]), schema
                    )
                return

        if "object" in ensure_list(schema.get("type")):
            yield from self._non_string_max_length(instance, mL)

    # pylint: disable=unused-argument, arguments-renamed
    def minLength(self, validator, mL, instance, schema):
        if validator.is_type(instance, "string"):
            if schema.get("format") == "json":
                try:
                    instance = json.loads(instance)
                except:  # noqa: E722
                    pass
                    return
                yield from self._non_string_min_length(instance, mL)
                return
            else:
                if len(instance) < mL:
                    yield ValidationError(f"{instance!r} is shorter than {mL}")
                return

        # there are scenarios where Fn::Sub may not predictable so use
        # best judgement
        key, value = is_function(instance)
        if key is not None:
            if key == "Fn::Sub":
                value = instance[key]
                if isinstance(value, str):
                    yield from self.minLength(
                        validator, mL, self._fix_sub_string(value), schema
                    )
                elif isinstance(value, list) and len(value) == 2:
                    yield from self.minLength(
                        validator, mL, self._fix_sub_string(value[0]), schema
                    )
                return

        if "object" in ensure_list(schema.get("type")):
            yield from self._non_string_min_length(instance, mL)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import logging
from collections import deque
from typing import Any

from cfnlint.helpers import FUNCTIONS, is_function
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema
from cfnlint.schema.manager import PROVIDER_SCHEMA_MANAGER

LOGGER = logging.getLogger(__name__)


class Properties(CfnLintJsonSchema):
    """Check Base Resource Configuration"""

    id = "E3002"
    shortdesc = "Resource properties are invalid"
    description = "Making sure that resources properties are properly configured"
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/cfn-schema-specification.md#properties"
    tags = ["resources"]

    def __init__(self):
        """Init"""
        super().__init__(
            keywords=["Resources/*"],
            all_matches=True,
        )
        self.rule_set = {
            "additionalProperties": "E3002",
            "anyOf": "E3017",
            "properties": "E3002",
            "dependentExcluded": "E3020",
            "dependentRequired": "E3021",
            "required": "E3003",
            "requiredOr": "E3058",
            "requiredXor": "E3014",
            "enum": "E3030",
            "enumCaseInsensitive": "E3030",
            "type": "E3012",
            "minLength": "E3033",
            "maxLength": "E3033",
            "uniqueItems": "E3037",
            "maximum": "E3034",
            "minimum": "E3034",
            "exclusiveMaximum": "E3034",
            "exclusiveMinimum": "E3034",
            "maxItems": "E3032",
            "minItems": "E3032",
            "pattern": "E3031",
            "prefixItems": "E3008",
            "oneOf": "E3018",
            "cfnLint": "E1101",
            "tagging": "E3024",
        }
        self.child_rules = dict.fromkeys(list(self.rule_set.values()))

    def validate(
        self, validator: Validator, _, instance: Any, schema: Any
    ) -> ValidationResult:
        if not validator.is_type(instance, "object"):
            return

        resolved_conditions = {}
        if validator.is_type(instance.get("Condition"), "string"):
            resolved_conditions = {instance.get("Condition"): True}

        validator = validator.evolve(
            context=validator.context.evolve(
                functions=list(FUNCTIONS),
                strict_types=False,
                conditions=validator.context.conditions.evolve(
                    resolved_conditions,
                ),
            ),
            function_filter=validator.function_filter.evolve(
                add_cfn_lint_keyword=True,
            ),
        )

        t = instance.get("Type")
        if not validator.is_type(t, "string"):
            return

        properties = instance.get("Properties", {})
        fn_k, fn_v = is_function(properties)
        if fn_k == "Ref" and fn_v == "AWS::NoValue":
            yield ValidationError(
                # Expected an object, received {"Ref":"AWS::NoValue"}
                message=f"{properties!r} is not of type object",
                path=deque(["Properties", fn_k]),
                rule=self.child_rules.get(self.rule_set.get("type")),  # type: ignore
                validator="type",
            )
            return

        for regions, schema in PROVIDER_SCHEMA_MANAGER.get_resource_schemas_by_regions(
            t, validator.context.regions
        ):
            region_validator = validator.evolve(
                context=validator.context.evolve(
                    regions=regions,
                    path=validator.context.path.evolve(
                        cfn_path=deque(["Resources", t, "Properties"]),
                    ).descend(path="Properties"),
                ),
            )

            region_validator = self.extend_validator(
                region_validator, schema.schema, region_validator.context.evolve()
            )
            for err in self._validate(region_validator, properties):
                err.path.appendleft("Properties")
                yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json
import logging
import os
from test.testlib.testcase import BaseTestCase
from unittest.mock import MagicMock, mock_open, patch

from cfnlint.schema._patch import SchemaPatch
from cfnlint.schema.other_schema_manager import OtherSchemaManager, SchemaNotFoundError

LOGGER = logging.getLogger("cfnlint.schema.other_schema_manager")
LOGGER.disabled = True


class TestOtherSchemaManager(BaseTestCase):
    """Used for Testing Other Schema Manager"""

    def setUp(self) -> None:
        super().setUp()
        self.manager = OtherSchemaManager()
        self.test_schema = {
            "type": "object",
            "properties": {"test": {"type": "string"}},
        }
        self.schema_patch = [
            {"op": "add", "path": "/additionalProperties", "value": False}
        ]

    @patch("cfnlint.schema.other_schema_manager.load_resource")
    def test_get_schema(self, mock_load_resource):
        """Test getting a schema"""
        mock_load_resource.return_value = self.test_schema

        # First call should load from resource
        schema = self.manager.get_schema("other.functions.join")
        self.assertEqual(schema, self.test_schema)
        mock_load_resource.assert_called_once()

        # Second call should use cached value
        schema = self.manager.get_schema("other.functions.join")
        self.assertEqual(schema, self.test_schema)
        mock_load_resource.assert_called_once()  # Still only called once

    @patch("cfnlint.schema.other_schema_manager.load_resource")
    def test_get_schema_not_found(self, mock_load_resource):
        """Test getting a schema that doesn't exist"""
        mock_load_resource.side_effect = Exception("Schema not found")

        with self.assertRaises(SchemaNotFoundError):
            self.manager.get_schema("nonexistent.schema")

    @patch("cfnlint.schema.other_schema_manager.load_resource")
    def test_reset(self, mock_load_resource):
        """Test resetting the cache"""
        mock_load_resource.return_value = self.test_schema

        # Load schema into cache
        self.manager.get_schema("other.functions.join")
        mock_load_resource.assert_called_once()

        # Reset cache
        self.manager.reset()

        # Should load from resource again
        self.manager.get_schema("other.functions.join")
        self.assertEqual(mock_load_resource.call_count, 2)

    @patch("cfnlint.schema.other_schema_manager.load_resource")
    @patch("cfnlint.schema.other_schema_manager.jsonpatch.JsonPatch")
    def test_patch(self, mock_json_patch, mock_load_resource):
        """Test patching a schema"""
        mock_load_resource.return_value = self.test_schema
        mock_patch = MagicMock()
        mock_json_patch.return_value = mock_patch

        # Load schema
        self.manager.get_schema("other.functions.join")

        # Create patch
        patch = SchemaPatch([], [], {"other.functions.join": self.schema_patch})

        # Apply patch
        self.manager.patch(patch, "us-east-1")

        # Verify patch was applied
        mock_json_patch.assert_called_with(self.schema_patch)
        mock_patch.apply.assert_called_once()

    @patch("cfnlint.schema.other_schema_manager.load_resource")
    @patch("cfnlint.schema.other_schema_manager.jsonpatch.JsonPatch")
    def test_patch_schema_not_found(self, mock_json_patch, mock_load_resource):
        """Test patching a schema that doesn't exist"""
        mock_load_resource.side_effect = Exception("Schema not found")
        mock_patch = MagicMock()
        mock_json_patch.return_value = mock_patch

        # Create patch for non-existent schema
        patch = SchemaPatch([], [], {"nonexistent.schema": self.schema_patch})

        # Apply patch - should not raise an exception
        self.manager.patch(patch, "us-east-1")

        # Verify patch was not applied
        mock_patch.apply.assert_not_called()

    @patch("cfnlint.schema.other_schema_manager.load_resource")
    @patch("cfnlint.schema.other_schema_manager.jsonpatch.JsonPatch")
    @patch("cfnlint.schema.other_schema_manager.sys.exit")
    @patch("cfnlint.schema.other_schema_manager.print")
    def test_patch_failure(
        self, mock_print, mock_exit, mock_json_patch, mock_load_resource
    ):
        """Test when patching a schema fails"""
        # Setup the schema to be loaded successfully
        mock_load_resource.return_value = self.test_schema

        # Setup the JsonPatch to raise an exception when apply is called
        mock_patch = MagicMock()
        mock_patch.apply.side_effect = Exception("Invalid patch operation")
        mock_json_patch.return_value = mock_patch

        # Create a patch
        schema_path = "other.functions.join"
        patch = SchemaPatch([], [], {schema_path: self.schema_patch})

        # Apply the patch
        self.manager.patch(patch, "us-east-1")

        # Verify that print was called with the error message
        mock_print.assert_called_with(
            (
                f"Error applying patch {self.schema_patch} for "
                f"{schema_path}: Invalid patch operation"
            )
        )

        # Verify that sys.exit was called with exit code 1
        mock_exit.assert_called_with(1)

    @patch("cfnlint.schema.other_schema_manager.open", new_callable=mock_open)
    @patch("cfnlint.schema.other_schema_manager.os.walk")
    @patch("cfnlint.schema.other_schema_manager.os.path.join")
    def test_load_registry_schemas(self, mock_path_join, mock_walk, mock_open_file):
        """Test loading registry schemas"""
        # Setup mock for os.walk to return one file
        mock_walk.return_value = [
            (os.path.normpath("/path/to/schemas"), [], ["test.json"])
        ]

        # Setup mock for os.path.join to return platform-appropriate paths
        mock_path_join.return_value = os.path.normpath("/path/to/schemas/test.json")

        # Setup mock for open to return a file with test schema
        mock_open_file.return_value.__enter__.return_value.read.return_value = (
            json.dumps(self.test_schema)
        )

        # Mock json.load to return our test schema
        with patch(
            "cfnlint.schema.other_schema_manager.json.load",
            return_value=self.test_schema,
        ):
            self.manager.load_registry_schemas(os.path.normpath("/path/to/schemas"))

        # Verify schema was loaded
        self.assertEqual(
            self.manager._schemas[os.path.normpath("/path/to/schemas")],
            self.test_schema,
        )

        # Verify open was called with correct path
        mock_open_file.assert_called_with(
            os.path.normpath("/path/to/schemas/test.json"), "r", encoding="utf-8"
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any, Sequence

import regex as re

from cfnlint.helpers import ensure_list, is_types_compatible
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.functions._BaseFn import BaseFn, all_types
from cfnlint.schema import PROVIDER_SCHEMA_MANAGER


class GetAtt(BaseFn):
    """Check if GetAtt values are correct"""

    id = "E1010"
    shortdesc = "GetAtt validation of parameters"
    description = (
        "Validates that GetAtt parameters are to valid resources and properties of"
        " those resources"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-getatt.html"
    tags = ["functions", "getatt"]

    def __init__(self) -> None:
        super().__init__(
            "Fn::GetAtt",
            all_types,
        )

    def _resolve_getatt(
        self,
        validator: Validator,
        key: str,
        value: Any,
        instance: Any,
        s: Any,
        paths: Sequence[Any],
    ) -> ValidationResult:

        for resource_name, resource_name_validator, _ in validator.resolve_value(
            value[0]
        ):
            for err in self.fix_errors(
                resource_name_validator.descend(
                    resource_name,
                    {"enum": list(validator.context.resources.keys())},
                    path=key,
                )
            ):
                err.path.append(paths[0])
                if err.instance != value[0]:
                    err.message = err.message + f" when {value[0]!r} is resolved"
                yield err
                break
            else:
                t = validator.context.resources[resource_name].type
                for (
                    regions,
                    schema,
                ) in PROVIDER_SCHEMA_MANAGER.get_resource_schemas_by_regions(
                    t, validator.context.regions
                ):
                    region = regions[0]
                    for attribute_name, _, _ in validator.resolve_value(value[1]):
                        if all(
                            not (bool(re.fullmatch(each, attribute_name)))
                            for each in validator.context.resources[
                                resource_name
                            ].get_atts(region)
                        ):
                            err = ValidationError(
                                (
                                    f"{attribute_name!r} is not one of "
                                    f"{validator.context.resources[resource_name].get_atts(region)!r}"
                                    f" in {regions!r}"
                                ),
                                validator=self.fn.py,
                                path=deque([self.fn.name, 1]),
                            )
                            if attribute_name != value[1]:
                                err.message = (
                                    err.message + f" when {value[1]!r} is resolved"
                                )
                            yield err
                            continue

                        evolved = validator.evolve(schema=s)  # type: ignore
                        evolved.validators = {  # type: ignore
                            "type": validator.validators.get("type"),  # type: ignore
                        }

                        getatts = validator.cfn.get_valid_getatts()
                        t = validator.context.resources[resource_name].type
                        pointer = getatts.match(region, [resource_name, attribute_name])

                        getatt_schema = schema.resolver.resolve_cfn_pointer(pointer)
                        # there is one exception we need to handle.  The resource type
                        # has a mix of types the input is integer and the output
                        # is string.  Since this is the only occurence
                        # we are putting in an exception to it.
                        if (
                            validator.context.resources[resource_name].type
                            == "AWS::DocDB::DBCluster"
                            and attribute_name == "Port"
                        ):
                            getatt_schema = {"type": "string"}

                        if not getatt_schema.get("type") or not s.get("type"):
                            continue

                        schema_types = ensure_list(getatt_schema.get("type"))
                        types = ensure_list(s.get("type"))

                        # GetAtt type checking is strict.
                        # It must match in all cases
                        if any(t in ["boolean", "integer", "boolean"] for t in types):
                            # this should be switched to validate the value of the
                            # property if it was available
                            continue
                        if is_types_compatible(types, schema_types, True):
                            continue

                        reprs = ", ".join(repr(type) for type in types)
                        yield ValidationError(
                            (f"{instance!r} is not of type {reprs}"),
                            validator=self.fn.py,
                            path=deque([self.fn.name]),
                            schema_path=deque(["type"]),
                        )

    def fn_getatt(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> ValidationResult:
        errs = list(super().validate(validator, s, instance, schema))
        if errs:
            yield from iter(errs)
            return

        key, value = self.key_value(instance)
        paths: list[int | None] = [0, 1]
        if validator.is_type(value, "string"):
            paths = [None, None]
            value = value.split(".", 1)

        errs = list(
            self._resolve_getatt(
                self.validator(validator, schema), key, value, instance, s, paths
            )
        )
        if errs:
            yield from iter(errs)
            return

        keyword = validator.context.path.cfn_path_string
        for rule in self.child_rules.values():
            if rule is None:
                continue
            if keyword in rule.keywords or "*" in rule.keywords:  # type: ignore
                yield from rule.validate(validator, s, value, s)  # type: ignore



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.integration import BaseCliTestCase

from cfnlint import ConfigMixIn


class TestQuickStartTemplates(BaseCliTestCase):
    """Test QuickStart Templates Parsing"""

    scenarios = [
        {
            "filename": "test/fixtures/templates/quickstart/nist_high_main.yaml",
            "results_filename": (
                "test/fixtures/results/quickstart/non_strict/nist_high_main.json"
            ),
            "exit_code": 14,
        },
        {
            "filename": "test/fixtures/templates/quickstart/nist_application.yaml",
            "results_filename": (
                "test/fixtures/results/quickstart/non_strict/nist_application.json"
            ),
            "exit_code": 12,
        },
        {
            "filename": "test/fixtures/templates/quickstart/openshift.yaml",
            "results_filename": (
                "test/fixtures/results/quickstart/non_strict/openshift.json"
            ),
            "exit_code": 12,
        },
        {
            "filename": "test/fixtures/templates/quickstart/cis_benchmark.yaml",
            "results_filename": (
                "test/fixtures/results/quickstart/non_strict/cis_benchmark.json"
            ),
            "exit_code": 4,
        },
    ]

    def test_templates(self):
        """Test Successful JSON Parsing"""
        self.maxDiff = None
        self.run_scenarios(
            [
                "--include-checks",
                "I",
                "--include-experimental",
                "--configure-rule",
                "E3012:strict=false",
            ]
        )

    def test_module_integration(self):
        """Test same templates using integration approach"""
        self.run_module_integration_scenarios(
            ConfigMixIn(
                [],
                include_checks=["I"],
                configure_rules={
                    "E3012": {
                        "strict": False,
                    }
                },
                include_experimental=True,
            )
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from cfnlint.rules.formats.FormatKeyword import FormatKeyword


class SecurityGroupId(FormatKeyword):
    id = "E1150"
    shortdesc = "Validate security group format"
    description = (
        "Security groups have to ref/gettatt to a security "
        "group or has the valid pattern"
    )
    tags = []
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/format_keyword.md#AWS::EC2::SecurityGroup.Id"

    def __init__(self):
        super().__init__(
            format="AWS::EC2::SecurityGroup.Id",
            pattern=r"^sg-([a-fA-F0-9]{8}|[a-fA-F0-9]{17})$",
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import json
from collections import deque
from copy import deepcopy
from typing import Any

import cfnlint.data.schemas.other.step_functions
from cfnlint.helpers import is_function
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.helpers import get_value_from_path
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails
from cfnlint.schema.resolver import RefResolver


class StateMachineDefinition(CfnLintJsonSchema):
    id = "E3601"
    shortdesc = "Validate the structure of a StateMachine definition"
    description = (
        "Validate the Definition or DefinitionString inside a "
        "AWS::StepFunctions::StateMachine resource"
    )
    source_url = "https://docs.aws.amazon.com/step-functions/latest/dg/amazon-states-language-state-machine-structure.html"
    tags = ["resources", "statemachine"]

    def __init__(self):
        super().__init__(
            keywords=[
                "Resources/AWS::StepFunctions::StateMachine/Properties",
                # https://github.com/aws-cloudformation/cfn-lint/issues/3518
                # Resources/AWS::StepFunctions::StateMachine/Properties/DefinitionString
            ],
            schema_details=SchemaDetails(
                cfnlint.data.schemas.other.step_functions, "statemachine.json"
            ),
            all_matches=True,
        )

        self.store = {
            "definition": self.schema,
        }

        self.resolver = RefResolver.from_schema(self.schema, store=self.store)

    def _fix_message(self, err: ValidationError) -> ValidationError:
        if len(err.path) > 1:
            err.message = f"{err.message} at {'/'.join(err.path)!r}"
        for i, c_err in enumerate(err.context):
            err.context[i] = self._fix_message(c_err)
        return err

    def _convert_schema_to_jsonata(self):
        schema = self.schema
        schema = deepcopy(schema)
        schema["definitions"]["common"] = deepcopy(schema["definitions"]["common"])
        schema["definitions"]["common"]["allOf"] = [
            {
                "properties": {
                    "QueryLanguage": {"const": "JSONata"},
                    "InputPath": False,
                    "OutputPath": False,
                    "Parameters": False,
                    "ResultPath": False,
                    "ResultSelector": False,
                },
            },
        ]
        schema["definitions"]["choice"]["allOf"] = [
            {
                "properties": {
                    "Choices": {
                        "items": {
                            "properties": {
                                "Condition": {"type": "string"},
                                "Next": {"pattern": "^.{1,128}$", "type": "string"},
                                "Assign": {"type": "object"},
                            },
                            "additionalProperties": False,
                            "required": ["Condition", "Next"],
                            "type": "object",
                        }
                    }
                }
            }
        ]
        for k in ["ItemsPath", "MaxConcurrencyPath"]:
            schema["definitions"]["map"]["properties"][k] = False
        return schema

    def _clean_schema(self, validator: Validator, instance: Any):
        for ql, ql_validator in get_value_from_path(
            validator, instance, deque(["QueryLanguage"])
        ):
            if ql is None or ql == "JSONPath":
                yield self.schema, ql_validator

            if ql == "JSONata":
                yield self._convert_schema_to_jsonata(), ql_validator

    def _validate_step(
        self,
        validator: Validator,
        substitutions: Any,
        value: Any,
        add_path_to_message: bool,
        k: str,
    ) -> ValidationResult:

        for err in validator.iter_errors(value):
            if validator.is_type(err.instance, "string"):
                if (
                    err.instance.replace("${", "").replace("}", "").strip()
                    in substitutions
                ):
                    continue
            if add_path_to_message:
                err = self._fix_message(err)

            err.path.appendleft(k)
            if err.schema in [True, False]:
                err.message = (
                    f"Additional properties are not allowed ({err.path[-1]!r} "
                    "was unexpected)"
                )
            if not err.validator or (
                not err.validator.startswith("fn_") and err.validator not in ["cfnLint"]
            ):
                err.rule = self

            yield self._clean_error(err)

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        definition_keys = ["Definition"]
        if not validator.cfn.has_serverless_transform():
            definition_keys.append("DefinitionString")

        substitutions = []
        props_substitutions = instance.get("DefinitionSubstitutions", {})
        if validator.is_type(props_substitutions, "object"):
            substitutions = list(props_substitutions.keys())

        # First time child rules are configured against the rule
        # so we can run this now
        for k in definition_keys:
            value = instance.get(k)
            fn_name, _ = is_function(value)
            if fn_name:
                continue
            if not value:
                continue

            add_path_to_message = False
            if validator.is_type(value, "string"):
                try:
                    value = json.loads(value)
                    add_path_to_message = True
                    for schema, schema_validator in self._clean_schema(
                        validator, value
                    ):
                        resolver = RefResolver.from_schema(schema, store=self.store)
                        step_validator = schema_validator.evolve(
                            context=validator.context.evolve(
                                functions=[],
                            ),
                            resolver=resolver,
                            schema=self.schema,
                        )

                        yield from self._validate_step(
                            step_validator, substitutions, value, add_path_to_message, k
                        )
                except json.JSONDecodeError:
                    return
            else:
                for schema, schema_validator in self._clean_schema(validator, value):
                    resolver = RefResolver.from_schema(schema, store=self.store)
                    step_validator = schema_validator.evolve(
                        resolver=resolver,
                        schema=schema,
                    )
                    yield from self._validate_step(
                        step_validator, substitutions, value, add_path_to_message, k
                    )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from cfnlint.rules import CloudFormationLintRule


class Required(CloudFormationLintRule):
    """Check Required Resource Configuration"""

    id = "E3003"
    shortdesc = "Required Resource properties are missing"
    description = "Make sure that Resources properties that are required exist"
    source_url = "https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/cfn-schema-specification.md#required"
    tags = ["resources", "properties", "required"]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.events.RuleScheduleExpression import RuleScheduleExpression


@pytest.fixture(scope="module")
def rule():
    rule = RuleScheduleExpression()
    yield rule


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "wrong format",
            [],
            [],
        ),
        (
            "10:15 AM (UTC) every day",
            "cron(15 10 * * ? *)",
            [],
        ),
        (
            "6:00 PM Monday through Friday",
            "cron(0 18 ? * MON-FRI *)",
            [],
        ),
        (
            "8:00 AM on the first day of the month",
            "cron(0 8 1 * ? *)",
            [],
        ),
        (
            "Every 10 min on weekdays",
            "cron(0/10 * ? * MON-FRI *)",
            [],
        ),
        (
            "Every 5 minutes between 8:00 AM and 5:55 PM weekdays",
            "cron(0/5 8-17 ? * MON-FRI *)",
            [],
        ),
        (
            "9:00 AM on the first Monday of each month",
            "cron(0 9 ? * 2#1 *)",
            [],
        ),
        (
            "Every 5 minutes",
            "rate(5 minutes)",
            [],
        ),
        (
            "Every hour",
            "rate(1 hour)",
            [],
        ),
        (
            "Every seven days",
            "rate(7 days)",
            [],
        ),
        (
            "has to be cron() or rate()",
            "daily",
            [
                ValidationError(
                    ("'daily' has to be either 'cron()' or 'rate()'"),
                )
            ],
        ),
        (
            "Empty cron",
            "cron()",
            [
                ValidationError(
                    ("'' is not of type 'string'"),
                )
            ],
        ),
        (
            "Empty rate",
            "rate()",
            [
                ValidationError(
                    ("'' is not of type 'string'"),
                )
            ],
        ),
        (
            "Not enough values",
            "rate(5)",
            [
                ValidationError(
                    ("'5' has to be of format rate(Value Unit)"),
                )
            ],
        ),
        (
            "Value has to be a digit",
            "rate(five minutes)",
            [
                ValidationError(
                    ("'five' is not of type 'integer'"),
                )
            ],
        ),
        (
            "Not enough values",
            "cron(0 */1 * * WED)",
            [
                ValidationError(
                    (
                        (
                            "'0' is not of length 6. (Minutes Hours "
                            "Day-of-month Month Day-of-week Year)"
                        )
                    ),
                )
            ],
        ),
        (
            (
                "Specify the Day-of-month and Day-of-week "
                "fields in the same cron expression"
            ),
            "cron(* 1 * * * *)",
            [
                ValidationError(
                    (
                        (
                            "'*' specifies both Day-of-month and "
                            "Day-of-week. (Minutes Hours "
                            "Day-of-month Month Day-of-week Year)"
                        )
                    ),
                )
            ],
        ),
        (
            "Value of 1 should be singular. 'minute' not 'minutes'",
            "rate(1 minutes)",
            [
                ValidationError(
                    ("'minutes' is not one of ['minute', 'hour', 'day']"),
                )
            ],
        ),
        (
            "Value has to be greater than 0",
            "rate(0 hour)",
            [
                ValidationError(
                    ("'0' is less than the minimum of 0"),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, expected, rule, validator):

    errs = list(rule.validate(validator, {}, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.context import Context
from cfnlint.context.context import Parameter, _init_parameters


@pytest.mark.parametrize(
    "name,instance,expected_type,expected_ref_value",
    [
        ("Valid string parameter", {"Type": "string"}, "string", []),
        (
            "Valid parameter with default and allowed values",
            {
                "Type": "String",
                "Default": "foo",
                "AllowedValues": [
                    "foo",
                    "bar",
                ],
            },
            "String",
            [
                ("foo", deque(["AllowedValues", 0])),
                ("bar", deque(["AllowedValues", 1])),
            ],
        ),
        (
            "Valid parameter with a CommaDelimitedList",
            {
                "Type": "CommaDelimitedList",
                "Default": "foo",
            },
            "CommaDelimitedList",
            [
                (["foo"], deque(["Default"])),
            ],
        ),
        (
            "Valid parameter with a List of Numbers",
            {
                "Type": "List<Number>",
                "Default": "10,20",
            },
            "List<Number>",
            [(["10", "20"], deque(["Default"]))],
        ),
        (
            "Valid parameter with a List of numbers for allowed values",
            {
                "Type": "List<Number>",
                "AllowedValues": ["10,20"],
            },
            "List<Number>",
            [(["10", "20"], deque(["AllowedValues", 0]))],
        ),
        (
            "Valid parameter with a SSM Parameter",
            {
                "Type": "AWS::SSM::Parameter::Value<String>",
                "Default": "foo",
            },
            "AWS::SSM::Parameter::Value<String>",
            [],
        ),
        (
            "Valid parameter with a MinValue and MaxValue",
            {"Type": "Number", "MinValue": "10", "MaxValue": "20"},
            "Number",
            [("10", deque(["MinValue"])), ("20", deque(["MaxValue"]))],
        ),
        (
            "Valid list parameter with an integer value",
            {"Type": "List<Number>", "Default": 10},
            "List<Number>",
            [(["10"], deque(["Default"]))],
        ),
        (
            "Valid list parameter with an integer value",
            {"Type": "List<Number>", "AllowedValues": [10]},
            "List<Number>",
            [(["10"], deque(["AllowedValues", 0]))],
        ),
    ],
)
def test_parameter(name, instance, expected_type, expected_ref_value):
    context = Context(["us-east-1"])
    parameter = Parameter(instance)

    assert expected_type == parameter.type
    assert expected_ref_value == list(
        parameter.ref_value(context)
    ), f"{name!r} test got {list(parameter.ref_value(context))}"
    assert {} == parameter.ref()


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        ("Valid string parameter", {"Type": "string"}, False),
        (
            "Valid string parameter with no echo string",
            {"Type": "string", "NoEcho": "true"},
            True,
        ),
        (
            "Valid string parameter with no echo boolean",
            {"Type": "string", "NoEcho": True},
            True,
        ),
    ],
)
def test_no_echo(name, instance, expected):
    parameter = Parameter(instance)

    assert expected == parameter.no_echo, f"{name} failed got {parameter.no_echo}"


@pytest.mark.parametrize(
    "name,instance",
    [
        ("Invalid Type", {"Type": {}}),
    ],
)
def test_errors(name, instance):
    with pytest.raises(ValueError):
        Parameter(instance)


def test_parameters():
    with pytest.raises(ValueError):
        _init_parameters([])



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.ectwo.PrivateIpWithNetworkInterface import (
    PrivateIpWithNetworkInterface,
)


@pytest.fixture(scope="module")
def rule():
    rule = PrivateIpWithNetworkInterface()
    yield rule


@pytest.mark.parametrize(
    "name,instance,path,expected",
    [
        (
            "Valid with no Private Ip Address",
            {
                "NetworkInterfaces": [
                    {
                        "PrivateIpAddresses": [
                            {"PrivateIpAddress": "172.31.35.42", "Primary": True}
                        ]
                    }
                ]
            },
            {
                "path": ["Resources", "Instance", "Properties"],
            },
            [],
        ),
        (
            "Valid with Private Ip Address with Primary False",
            {
                "PrivateIpAddress": "172.31.35.42",
                "NetworkInterfaces": [
                    {
                        "PrivateIpAddresses": [
                            {"PrivateIpAddress": "172.31.35.42", "Primary": False}
                        ]
                    }
                ],
            },
            {
                "path": ["Resources", "Instance", "Properties"],
            },
            [],
        ),
        (
            "Valid with Private Ip Address without Primary specified",
            {
                "PrivateIpAddress": "172.31.35.42",
                "NetworkInterfaces": [
                    {"PrivateIpAddresses": [{"PrivateIpAddress": "172.31.35.42"}]}
                ],
            },
            {
                "path": ["Resources", "Instance", "Properties"],
            },
            [],
        ),
        (
            "Valid AWS::EC2::NetworkInterface with PrivateIpAddresses",
            {
                "PrivateIpAddress": "172.31.35.42",
                "PrivateIpAddresses": [
                    {"PrivateIpAddress": "172.31.35.42", "Primary": False}
                ],
            },
            {
                "path": ["Resources", "Instance", "Properties"],
            },
            [],
        ),
        (
            "Invalid with a private ip address in two spots",
            {
                "PrivateIpAddress": "172.31.35.42",
                "NetworkInterfaces": [
                    {
                        "PrivateIpAddress": "172.31.35.42",
                    }
                ],
            },
            {
                "path": ["Resources", "Instance", "Properties"],
            },
            [
                ValidationError(
                    "'Primary' cannot be True when 'PrivateIpAddress' is specified",
                    validator=None,
                    rule=PrivateIpWithNetworkInterface(),
                    path=deque(["NetworkInterfaces", 0, "PrivateIpAddress"]),
                    schema_path=deque(
                        [
                            "then",
                            "properties",
                            "NetworkInterfaces",
                            "items",
                            "properties",
                            "PrivateIpAddress",
                        ]
                    ),
                )
            ],
        ),
        (
            "Invalid with a private ip address",
            {
                "PrivateIpAddress": "172.31.35.42",
                "NetworkInterfaces": [
                    {
                        "PrivateIpAddresses": [
                            {"PrivateIpAddress": "172.31.35.42", "Primary": True}
                        ]
                    }
                ],
            },
            {
                "path": ["Resources", "Instance", "Properties"],
            },
            [
                ValidationError(
                    "'Primary' cannot be True when 'PrivateIpAddress' is specified",
                    validator="enum",
                    rule=PrivateIpWithNetworkInterface(),
                    path=deque(
                        ["NetworkInterfaces", 0, "PrivateIpAddresses", 0, "Primary"]
                    ),
                    schema_path=deque(
                        [
                            "then",
                            "properties",
                            "NetworkInterfaces",
                            "items",
                            "properties",
                            "PrivateIpAddresses",
                            "items",
                            "properties",
                            "Primary",
                            "enum",
                        ]
                    ),
                )
            ],
        ),
        (
            "Invalid AWS::EC2::NetworkInterface with PrivateIpAddresses",
            {
                "PrivateIpAddress": "172.31.35.42",
                "PrivateIpAddresses": [
                    {"PrivateIpAddress": "172.31.35.42", "Primary": True}
                ],
            },
            {
                "path": ["Resources", "Instance", "Properties"],
            },
            [
                ValidationError(
                    "'Primary' cannot be True when 'PrivateIpAddress' is specified",
                    validator="enum",
                    rule=PrivateIpWithNetworkInterface(),
                    path=deque(["PrivateIpAddresses", 0, "Primary"]),
                    schema_path=deque(
                        [
                            "then",
                            "properties",
                            "PrivateIpAddresses",
                            "items",
                            "properties",
                            "Primary",
                            "enum",
                        ]
                    ),
                )
            ],
        ),
    ],
    indirect=["path"],
)
def test_validate(name, instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert (
        errs == expected
    ), f"Expected test {name!r} to have {expected!r} but got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.functions.Length import Length


@pytest.fixture
def rule():
    rule = Length()
    yield rule


@pytest.mark.parametrize(
    "name,instance,schema,template,expected",
    [
        (
            "Fn::Length is not supported",
            {"Fn::Length": []},
            {"type": "integer"},
            {},
            [
                ValidationError(
                    (
                        "Fn::Length is not supported without "
                        "'AWS::LanguageExtensions' transform"
                    ),
                    path=deque([]),
                    schema_path=deque([]),
                    validator="fn_length",
                    rule=Length(),
                ),
            ],
        ),
        (
            "Fn::Length valid structure",
            {"Fn::Length": []},
            {"type": "integer"},
            {"Transform": ["AWS::LanguageExtensions"]},
            [],
        ),
        (
            "Fn::Length invalid type",
            {"Fn::Length": "foo"},
            {"type": "integer"},
            {"Transform": ["AWS::LanguageExtensions"]},
            [
                ValidationError(
                    "'foo' is not of type 'array'",
                    path=deque(["Fn::Length"]),
                    schema_path=deque(["cfnContext", "schema", "type"]),
                    validator="fn_length",
                    rule=Length(),
                ),
            ],
        ),
        (
            "Fn::Length invalid output type",
            {"Fn::Length": ["foo"]},
            {"type": "array"},
            {"Transform": ["AWS::LanguageExtensions"]},
            [
                ValidationError(
                    "{'Fn::Length': ['foo']} is not of type 'array'",
                    path=deque([]),
                    schema_path=deque([]),
                    validator="fn_length",
                    rule=Length(),
                ),
            ],
        ),
        (
            "Fn::Length using valid function",
            {"Fn::Length": {"Fn::GetAZs": ""}},
            {"type": "integer"},
            {"Transform": ["AWS::LanguageExtensions"]},
            [],
        ),
        (
            "Fn::Length using valid functions in array",
            {"Fn::Length": [{"Ref": "MyResource"}]},
            {"type": "integer"},
            {"Transform": ["AWS::LanguageExtensions"]},
            [],
        ),
        (
            "Fn::Length is not supported",
            {"Fn::Length": []},
            {"type": "integer"},
            {"Transform": ["AWS::LanguageExtensions"]},
            [],
        ),
        (
            "Fn::Length output while a number can be a string",
            {"Fn::Length": []},
            {"type": "string"},
            {"Transform": ["AWS::LanguageExtensions"]},
            [],
        ),
    ],
    indirect=["template"],
)
def test_validate(name, instance, schema, expected, validator, rule):
    errs = list(rule.fn_length(validator, schema, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.rds.DbInstanceEngineVersion import DbInstanceEngineVersion


@pytest.fixture(scope="module")
def rule():
    rule = DbInstanceEngineVersion()
    yield rule


@pytest.fixture
def schema():
    return {
        "allOf": [
            {
                "if": {
                    "properties": {"Engine": {"type": "string"}},
                    "required": ["Engine"],
                },
                "then": {
                    "properties": {
                        "Engine": {
                            "enum": [
                                "mysql",
                            ]
                        }
                    }
                },
            },
            {
                "if": {
                    "properties": {
                        "Engine": {"const": "mysql"},
                        "EngineVersion": {"type": ["string", "number"]},
                    },
                    "required": ["Engine", "EngineVersion"],
                },
                "then": {
                    "properties": {
                        "EngineVersion": {
                            "enum": [
                                "8.0.39",
                                "8.0.40",
                                "8.0.41",
                                "8.4.3",
                                "8.4.4",
                            ]
                        }
                    }
                },
            },
        ],
    }


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "Engine": "mysql",
            },
            [],
        ),
        (
            [],
            [],
        ),
        (
            {
                "Engine": "MySqL",  # API converts it appropriately
            },
            [],
        ),
        (
            {"Engine": {"Ref": "Engine"}},
            [],
        ),
        (
            {"Engine": {"Ref": "Engine"}, "EngineVeresion": {"Ref": "EngineVersion"}},
            [],
        ),
        (
            {
                "Engine": "MySqL",  # API converts it appropriately
                "EngineVeresion": {"Ref": "EngineVersion"},
            },
            [],
        ),
        (
            {"Engine": "foo"},
            [
                ValidationError(
                    ("'foo' is not one of ['mysql']"),
                    rule=DbInstanceEngineVersion(),
                    path=deque(["Engine"]),
                    validator="enum",
                    schema_path=deque(
                        ["allOf", 0, "then", "properties", "Engine", "enum"]
                    ),
                )
            ],
        ),
        (
            {"Engine": "mysql", "EngineVersion": "foo"},
            [
                ValidationError(
                    (
                        "'foo' is not one of ['8.0.39', '8.0.40', "
                        "'8.0.41', '8.4.3', '8.4.4']"
                    ),
                    rule=DbInstanceEngineVersion(),
                    path=deque(["EngineVersion"]),
                    validator="enum",
                    schema_path=deque(
                        ["allOf", 1, "then", "properties", "EngineVersion", "enum"]
                    ),
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator, schema):
    rule._schema = schema
    errs = list(rule.validate(validator, "", instance, {}))
    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.other.resources as resources_schemas
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class UpdateReplacePolicyDeletionPolicy(CfnLintJsonSchema):
    """Check resources with UpdateReplacePolicy/DeletionPolicy have both"""

    id = "W3011"
    shortdesc = "Check resources with UpdateReplacePolicy/DeletionPolicy have both"
    description = (
        "Both UpdateReplacePolicy and DeletionPolicy are needed to protect resources"
        " from deletion"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-attribute-deletionpolicy.html"
    tags = ["resources", "updatereplacepolicy", "deletionpolicy"]

    def __init__(
        self,
    ) -> None:
        super().__init__(
            ["Resources/*"],
            SchemaDetails(resources_schemas, "update_policy_delete.json"),
            False,
        )

    def message(self, instance: Any, err: ValidationError) -> str:
        return (
            "Both 'UpdateReplacePolicy' and 'DeletionPolicy' "
            "are needed to protect resource from deletion"
        )

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        resource_type = instance.get("Type")
        if resource_type in [
            "AWS::Lambda::Version",
            "AWS::Lambda::LayerVersion",
        ]:
            return

        yield from super().validate(validator, keywords, instance, schema)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.lmbd.EventSourceMappingEventSourceArnSqsExclusive import (
    EventSourceMappingEventSourceArnSqsExclusive,
)


@pytest.fixture(scope="module")
def rule():
    rule = EventSourceMappingEventSourceArnSqsExclusive()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "EventSourceArn": "arn:aws:sqs:us-east-1:123456789012:sqs-arn",
            },
            [],
        ),
        (
            [],  # wrong type
            [],
        ),
        (
            {
                "EventSourceArn": "arn:aws:sqs:us-east-1:123456789012:sqs-arn",
                "StartingPosition": "1",
            },
            [
                ValidationError(
                    "Additional properties are not allowed ('StartingPosition')",
                    rule=EventSourceMappingEventSourceArnSqsExclusive(),
                    path=deque(["StartingPosition"]),
                    validator=None,
                    schema_path=deque(["then", "properties", "StartingPosition"]),
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.helpers import VALID_PARAMETER_TYPES, VALID_PARAMETER_TYPES_LIST
from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.rules.functions._BaseFn import BaseFn, all_types


class Ref(BaseFn):
    id = "E1020"
    shortdesc = "Ref validation of value"
    description = (
        "Making sure the Ref has a String value (no other functions are supported)"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-ref.html"
    tags = ["functions", "ref"]

    def __init__(self) -> None:
        super().__init__(
            "Ref",
            all_types,
            resolved_rule="W1030",
        )
        self._all_refs = [
            "W2010",
        ]
        self.child_rules.update(dict.fromkeys(self._all_refs))

    def ref(
        self,
        validator: Validator,
        subschema: Any,
        instance: Any,
        schema: dict[str, Any],
    ):
        yield from super().validate(validator, subschema, instance, schema)

        _, value = self.key_value(instance)
        if not validator.is_type(value, "string"):
            return

        if value in validator.context.parameters:
            parameter_type = validator.context.parameters[value].type
            schema_types = self.resolve_type(validator, subschema)
            if not schema_types:
                return
            reprs = ", ".join(repr(type) for type in schema_types)

            if all(
                st not in ["string", "boolean", "integer", "number"]
                for st in schema_types
            ):
                if parameter_type not in VALID_PARAMETER_TYPES_LIST:
                    yield ValidationError(f"{instance!r} is not of type {reprs}")
                    return
            elif all(st not in ["array"] for st in schema_types):
                if parameter_type not in [
                    x
                    for x in VALID_PARAMETER_TYPES
                    if x not in VALID_PARAMETER_TYPES_LIST
                ]:
                    yield ValidationError(f"{instance!r} is not of type {reprs}")
                    return

        for rule_id in self._all_refs:
            rule = self.child_rules.get(rule_id)
            if rule:
                yield from rule.validate(validator, {}, value, subschema)  # type: ignore[attr-defined]

        keyword = validator.context.path.cfn_path_string
        for rule in self.child_rules.values():
            if not rule or rule.id in self._all_refs:
                continue
            if not hasattr(rule, "keywords"):
                continue
            if keyword in rule.keywords or "*" in rule.keywords:  # type: ignore[attr-defined]
                yield from rule.validate(validator, keyword, value, subschema)  # type: ignore[attr-defined]



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.ecs.ServiceHealthCheckGracePeriodSeconds import (
    ServiceHealthCheckGracePeriodSeconds,
)


@pytest.fixture(scope="module")
def rule():
    rule = ServiceHealthCheckGracePeriodSeconds()
    yield rule


@pytest.fixture
def template():
    return {
        "Conditions": {
            "IsUsEast1": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]}
        },
    }


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {"HealthCheckGracePeriodSeconds": "Foo", "LoadBalancers": ["Bar"]},
            [],
        ),
        (
            {"LoadBalancers": []},
            [],
        ),
        (
            [],  # wrong type
            [],
        ),
        (
            {"HealthCheckGracePeriodSeconds": "Foo", "LoadBalancers": []},
            [
                ValidationError(
                    "expected minimum item count: 1, found: 0",
                    rule=ServiceHealthCheckGracePeriodSeconds(),
                    path=deque(["LoadBalancers"]),
                    validator="minItems",
                    schema_path=deque(
                        ["then", "then", "properties", "LoadBalancers", "minItems"]
                    ),
                )
            ],
        ),
        (
            {
                "HealthCheckGracePeriodSeconds": "Foo",
            },
            [
                ValidationError(
                    "'LoadBalancers' is a required property",
                    rule=ServiceHealthCheckGracePeriodSeconds(),
                    path=deque([]),
                    validator="required",
                    schema_path=deque(["then", "required"]),
                )
            ],
        ),
        (
            {
                "HealthCheckGracePeriodSeconds": "Foo",
                "LoadBalancers": [
                    {"Fn::If": ["IsUsEast1", "Bar", {"Ref": "AWS::NoValue"}]}
                ],
            },
            [
                ValidationError(
                    "expected minimum item count: 1, found: 0",
                    rule=ServiceHealthCheckGracePeriodSeconds(),
                    path=deque(["LoadBalancers"]),
                    validator="minItems",
                    schema_path=deque(
                        ["then", "then", "properties", "LoadBalancers", "minItems"]
                    ),
                )
            ],
        ),
        (
            {
                "HealthCheckGracePeriodSeconds": "Foo",
                "LoadBalancers": {
                    "Fn::If": ["IsUsEast1", ["Bar"], {"Ref": "AWS::NoValue"}]
                },
            },
            [
                ValidationError(
                    "'LoadBalancers' is a required property",
                    rule=ServiceHealthCheckGracePeriodSeconds(),
                    path=deque([]),
                    validator="required",
                    schema_path=deque(["then", "required"]),
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any, Iterator

import cfnlint.data.schemas.extensions.aws_sqs_queue
from cfnlint.helpers import bool_compare, is_function
from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.helpers import get_value_from_path
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class QueueDLQ(CfnLintJsonSchema):
    id = "E3502"
    shortdesc = "Validate SQS DLQ queues are the same type"
    description = (
        "SQS queues using DLQ have to have the destination "
        "queue as the same type (FIFO or standard)"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-sqs-queue.html"
    tags = ["resources", "sqs"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::SQS::Queue/Properties",
            ],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_sqs_queue,
                filename="properties.json",
            ),
            all_matches=True,
        )

    def _is_fifo_queue(
        self, validator: Validator, instance: Any
    ) -> Iterator[tuple[str, Validator]]:
        standard = "standard"
        fifo = "FIFO"

        if "FifoQueue" not in instance:
            yield standard, validator
            return

        for queue_type, queue_type_validator in get_value_from_path(
            validator=validator, instance=instance, path=deque(["FifoQueue"])
        ):
            yield (
                fifo if bool_compare(queue_type, True) else standard
            ), queue_type_validator

    def validate(
        self, validator: Validator, _: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        for queue_type, queue_type_validator in self._is_fifo_queue(
            validator=validator,
            instance=instance,
        ):
            queue_type_validator = queue_type_validator.evolve(
                context=queue_type_validator.context.evolve(
                    path=validator.context.path.evolve()
                )
            )

            for target, target_validator in get_value_from_path(
                queue_type_validator,
                instance,
                path=deque(["RedrivePolicy", "deadLetterTargetArn"]),
            ):
                k, v = is_function(target)
                if k != "Fn::GetAtt":
                    return

                if target_validator.is_type(v, "string"):
                    v = v.split(".")

                if len(v) < 1:
                    return

                dest_queue = validator.cfn.template.get("Resources", {}).get(v[0], {})

                if dest_queue.get("Type") != "AWS::SQS::Queue":
                    return

                for dest_queue_type, _ in self._is_fifo_queue(
                    target_validator,
                    instance=dest_queue.get("Properties", {}),
                ):
                    if queue_type != dest_queue_type:
                        yield ValidationError(
                            (
                                f"Source queue type {queue_type!r} does not "
                                f"match destination queue type {dest_queue_type!r}"
                            ),
                            rule=self,
                            path=target_validator.context.path.path,
                        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.elasticache.CacheClusterEngine import CacheClusterEngine


@pytest.fixture(scope="module")
def rule():
    rule = CacheClusterEngine()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {"Engine": "redis", "EngineVersion": 7.1},
            [],
        ),
        (
            {
                "Engine": "redis",
            },
            [],
        ),
        (
            {},
            [],
        ),
        (
            {
                "Engine": "redis",
                "EngineVersion": "7.1.0",
            },
            [
                ValidationError(
                    (
                        "'7.1.0' is not one of "
                        "['4.0.10', '5.0.6', "
                        "'6.0', '6.2', '7.0', '7.1']"
                    ),
                    validator="enum",
                    path=deque(["EngineVersion"]),
                    schema_path=[
                        "allOf",
                        2,
                        "then",
                        "properties",
                        "EngineVersion",
                        "enum",
                    ],
                    rule=CacheClusterEngine(),
                )
            ],
        ),
        (
            {
                "Engine": "oss-redis",
                "EngineVersion": "7.1",
            },
            [
                ValidationError(
                    ("'oss-redis' is not one of " "['memcached', 'redis', 'valkey']"),
                    validator="enum",
                    path=deque(["Engine"]),
                    schema_path=["allOf", 0, "then", "properties", "Engine", "enum"],
                    rule=CacheClusterEngine(),
                )
            ],
        ),
        (
            {
                "Engine": "redis",
                "EngineVersion": {"Ref": "AWS::AccountId"},
            },
            [],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright (c) 2013 Julian Berman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

SPDX-License-Identifier: MIT
"""

# Code is taken from jsonschema package and adapted CloudFormation use
# https://github.com/python-jsonschema/jsonschema
from __future__ import annotations

from typing import Any

import regex as re

import cfnlint.jsonschema._keywords as validators_standard
from cfnlint.helpers import (
    BOOLEAN_STRINGS,
    FUNCTION_FOR_EACH,
    FUNCTION_TRANSFORM,
    ensure_list,
    is_function,
)
from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.jsonschema._typing import V, ValidationResult

_ap_exception_fns = set([FUNCTION_TRANSFORM, FUNCTION_FOR_EACH])


def additionalProperties(
    validator: Validator, aP: Any, instance: Any, schema: Any
) -> ValidationResult:
    # is function will just return if one item is present
    # as this is the standard. We will handle exceptions below
    k, _ = is_function(instance)
    if k in validator.context.functions:
        return
    for err in validators_standard.additionalProperties(
        validator, aP, instance, schema
    ):
        # Some functions can exist at the same level
        # so we need to validate that if those functions are
        # currently supported by the context and are part of the
        # error

        # if the path is 0 just yield the error and return
        # this should never happen
        if not len(err.path) > 0:  # pragma: no cover
            yield err  # pragma: no cover
            return  # pragma: no cover

        for fn in list(_ap_exception_fns & set(validator.context.functions)):
            if re.fullmatch(fn, str(err.path[0])):  # type: ignore
                break
        else:
            yield err


def cfnContext(
    validator: Validator,
    s: Any,
    instance: Any,
    schema: Any,
) -> ValidationResult:

    context_parameters: dict[str, Any] = {}

    functions = s.get("functions")
    if functions is not None:
        if validator.is_type(functions, "object"):
            if "$ref" in functions:
                _, functions = validator.resolver.resolve(functions["$ref"])
            else:
                functions = []

    pseudo_parameters = s.get("pseudoParameters")
    if pseudo_parameters is not None:
        context_parameters["pseudo_parameters"] = set(pseudo_parameters)

    references = s.get("references")
    if references is not None:
        if "Resources" not in references:
            context_parameters["resources"] = {}

    if functions is not None:
        context_parameters["functions"] = functions

    cfn_validator = validator.evolve(
        context=validator.context.evolve(**context_parameters)
    )

    yield from cfn_validator.descend(
        instance=instance,
        schema=s.get("schema", {}),
        schema_path="schema",
    )


def dynamicValidation(
    validator: Validator, dV: Any, instance: Any, schema: dict[str, Any]
) -> ValidationResult:
    """
    Performs dynamic validation based on context.

    The dynamicValidation keyword supports:
    - context: The context source to validate against
               (parameters, conditions, resources, etc.)
    - transformCheck: Check if a specific transform exists
                      (returns true/false for if/then/else)
    - pathCheck: Validate based on the current path in the template
    """
    if not validator.is_type(dV, "object"):
        return

    # Handle transform check (for use with if/then/else)
    transform_check = dV.get("transformCheck")
    if transform_check is not None:
        if transform_check not in validator.context.transforms.transforms:
            yield ValidationError(
                (
                    f"Transform {transform_check!r} is required "
                    "but not present in the template"
                )
            )

    # Handle dynamic source validation
    context_source = dV.get("context")
    if context_source is not None:
        if context_source and isinstance(context_source, str):
            # Get the appropriate collection based on the context source
            collection = None
            if context_source == "conditions":
                collection = list(validator.context.conditions.conditions.keys())
            elif context_source == "mappings":
                collection = list(validator.context.mappings.maps.keys())
            elif context_source == "refs":
                collection = validator.context.refs

            if collection is not None:
                # Build a dynamic schema with an enum of valid values
                dynamic_schema = {"enum": collection}

                # Use descend to validate against the dynamic schema
                yield from validator.descend(
                    instance=instance,
                    schema=dynamic_schema,
                )

    path_check = dV.get("pathCheck")
    if path_check:
        current_path = "/".join(str(p) for p in validator.context.path.path)
        pattern_schema = {"pattern": f"^{path_check}.*$"}

        yield from validator.descend(
            instance=current_path,
            schema=pattern_schema,
        )


#####
# Type checks
#####
def _raw_type(validator: Validator, tS: Any, instance: Any) -> bool:
    if tS in ["object", "array", "null"] or validator.is_type(instance, "null"):
        return validator.is_type(instance, tS)
    if "string" == tS:
        if validator.is_type(instance, "object") or validator.is_type(
            instance, "array"
        ):
            return False
        return True
    if "number" == tS:
        if validator.is_type(instance, "boolean"):
            return False
        try:
            float(instance)
            return True
        except (ValueError, TypeError):
            return False
    if "integer" == tS:
        if validator.is_type(instance, "boolean"):
            return False
        try:
            int(instance)
            return True
        except (ValueError, TypeError):
            return False
    if "boolean" == tS:
        if validator.is_type(instance, "boolean"):
            return True
        if instance in list(BOOLEAN_STRINGS):
            return True
    return False


def _raw_type_strict(validator: Validator, tS: Any, instance: Any) -> bool:
    return validator.is_type(instance, tS)


# pylint: disable=unused-argument
def cfn_type(validator: Validator, tS: Any, instance: Any, schema: Any):
    """
    When evaluating a type in CloudFormation we have to account
    for the intrinsic functions that the values can represent
    (Ref, GetAtt, If, ...).  This will evaluate if the correct type
    is not found and the instance is an object with a function
    that we do our best to evaluate if that function represents the
    type we are looking for
    """
    if validator.context.strict_types:
        raw_type_fn = _raw_type_strict
    else:
        raw_type_fn = _raw_type
    tS = ensure_list(tS)
    if not any(raw_type_fn(validator, type, instance) for type in tS):
        reprs = ", ".join(repr(type) for type in tS)
        yield ValidationError(f"{instance!r} is not of type {reprs}")


cfn_validators: dict[str, V] = {
    "additionalProperties": additionalProperties,
    "cfnContext": cfnContext,
    "dynamicValidation": dynamicValidation,
    "type": cfn_type,
}



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.jsonschema import CfnTemplateValidator
from cfnlint.rules.conditions.EqualsIsUseful import EqualsIsUseful


@pytest.mark.parametrize(
    "name,instance,num_of_errors",
    [
        ("Equal string and integer", [1, "1"], 1),
        ("Equal string and boolean", [True, "true"], 1),
        ("Equal string and number", [1.0, "1.0"], 1),
        ("Not equal string and integer", [1, "1.1"], 1),
        ("Not equal string and boolean", [True, "True"], 1),
        ("No error on bad type", {"true": True}, 0),
        ("No error on bad length", ["a", "a", "a"], 0),
        ("No with string and account id", ["A", {"Ref": "AWS::AccountId"}], 0),
        ("No with string and account id", [{"Ref": "AWS::AccountId"}, "A"], 0),
    ],
)
def test_names(name, instance, num_of_errors):
    rule = EqualsIsUseful()
    validator = CfnTemplateValidator({})
    assert (
        len(list(rule.equals_is_useful(validator, {}, instance, {}))) == num_of_errors
    ), f"Expected {num_of_errors} errors for {name} and {instance}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from test.testlib.testcase import BaseTestCase

from cfnlint.decode import convert_dict
from cfnlint.decode.node import dict_node, list_node, str_node


class TestConvertDict(BaseTestCase):
    """Test Converting regular dicts into CFN Dicts"""

    def test_success_run(self):
        """Test success run"""
        obj = {"Key": {"SubKey": "Value"}, "List": [{"SubKey": "AnotherValue"}]}

        results = convert_dict(obj)
        self.assertTrue(isinstance(results, dict_node))
        self.assertTrue(isinstance(results.get("Key"), dict_node))
        self.assertTrue(isinstance(results.get("List")[0], dict_node))
        self.assertTrue(isinstance(results.get("List"), list_node))
        for k, _ in results.items():
            self.assertTrue(isinstance(k, str_node))



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.parameters.ValuePattern import ValuePattern as ParameterPattern
from cfnlint.rules.resources.properties.Pattern import Pattern


@pytest.fixture(scope="module")
def rule():
    rule = Pattern()
    rule.child_rules["W2031"] = ParameterPattern()
    yield rule


@pytest.fixture
def template():
    return {
        "Transform": ["AWS::Serverless-2016-10-31"],
        "Parameters": {
            "SSMParameter": {
                "Type": "AWS::SSM::Parameter::Value<String>",
                "Default": "foo",
            },
            "Parameter": {
                "Type": "String",
                "Default": "bar",
            },
        },
    }


@pytest.mark.parametrize(
    "name,instance,pattern,expected",
    [
        (
            "Valid because SSM parameter default value",
            "foo",
            "bar",
            [],
        ),
        (
            "Invalid because not the SSM parameter",
            "bar",
            "foo",
            [
                ValidationError(
                    message="'bar' does not match 'foo'",
                )
            ],
        ),
        (
            "Invalid an unrelated to the parameters",
            "foobar",
            "foofoo",
            [
                ValidationError(
                    message="'foobar' does not match 'foofoo'",
                )
            ],
        ),
    ],
)
def test_validate(name, instance, pattern, expected, rule, validator):
    errs = list(rule.pattern(validator, pattern, instance, {}))
    assert errs == expected, f"{name} got errors {errs!r}"
